//-------------------------------------------------------------------------------------------
// Copyright (c) 2008-2010, Institute of Computer Technology, University of Technology Vienna
// Authors:
//   Christoph Grimm, Jan Haase, Markus Damm, Jiong Ou,  
//   Joseph Wenninger, Peter Brunmayr, Muhammad Farooq
// Contact:
//   tuv-ams-lib@ict.tuwien.ac.at
// For the license please read LICENSE.txt
//-------------------------------------------------------------------------------------------

// NEW STUFF:
//
// 04/18/2008
//
// TDF -> CSY conversion
// calling the converterchannel-method "clock_TDF()" returns a sc_signal<bool>
// which ticks with the speed of the TDF-cluster on the writing side of the converter channel
// (with 50-50 duty cycles); at each rising edge, there's a new value on the incoming TDF signal
//
// DE / CSY -> TDF conversion
// The DE->TDF converter now checks, if it misses signal changes on behalf of the DE / CSY side,
// i.e. if the TDF side "undersamples" the DE / CSY signal. if "missed values" are detected, a
// warning is raised. As of now, this warning is always on; an option for switching it on/off
// will be included
//
// only two additional data types on the reading side now - makes the code easier



#ifndef AHES_CONVERTER_CONVERTERCHANNEL_H
#define AHES_CONVERTER_CONVERTERCHANNEL_H 1

#ifdef AHES_CONVERTERS_TRACE_CONV
  #define CONVERTERS_TRACE_CONV_DO(x) x
#else
  #define CONVERTERS_TRACE_CONV_DO(x)
#endif

#include <systemc.h>
#include <systemc-ams.h>
#include "converters.h"
#include <vector>

#include <iostream>
#include <sstream>
using namespace std;


namespace TUV_ams_lib {

// this is a helperclass which enables us to store the input signal within a single variable
// depending on the input MoC of the converter channel
// eg. an sc_signal<int> will be a ahes_signal<SC, int> etc.

template <MoC THEMOC, typename T>
class TUV_ams_lib_signal{
  public: TUV_ams_lib_signal(){}
  public: TUV_ams_lib_signal(int x){}
};

template <typename T>
class TUV_ams_lib_signal<SC,T>: public sc_signal<T>{
  public: TUV_ams_lib_signal(){}
  public: TUV_ams_lib_signal(int x){}
};

template <typename T>
class TUV_ams_lib_signal<TDF,T>: public sca_tdf::sca_signal<T>{
  public: TUV_ams_lib_signal(){}
  public: TUV_ams_lib_signal(int x){}
};

template <typename T>
class TUV_ams_lib_signal<FIFO,T>: public sc_fifo<T>{
  public: TUV_ams_lib_signal(){}
  public: TUV_ams_lib_signal(int size): sc_fifo<T>(size){}
};

template <typename T>
class TUV_ams_lib_signal<ELEC_VOLTAGE,T>: public sca_eln::sca_node{
  public: TUV_ams_lib_signal(){}
  public: TUV_ams_lib_signal(int x){}
};

template <typename T>
class TUV_ams_lib_signal<ELEC_CURRENT,T>: public sca_eln::sca_node{
  public: TUV_ams_lib_signal(){}
  public: TUV_ams_lib_signal(int x){}
};



/////////////////////////////////////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// The converterchannel-class takes up to five template parameters
// MOC_WRITE: the MoC of the writing side
// T_WRITE: The datatype of the writing port, which can also be used on the reading side
// T_READ1, T_READ2: The datatypes of the reading ports (optional)
// note: T_WRITE, T_READ1, T_READ2 have to be pairwise distinct!
// the defaults for T_READ1 and T_READ2 are disfunctional dummy-datatypes defined in converters.h
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

template <MoC MOC_WRITE, typename T_WRITE, typename T_READ1=impl::dummy<1>, typename T_READ2=impl::dummy<2> >
class converterchannel : public sc_core::sc_channel{

  public:

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class data ///////////////////////////////////////////////////////////////////////////////////////////////

  sc_signal<bool>* tdf_clk;	// clock-signal which runs with the speed of the TDF input cluster (if any)
  tdf_clock* tdf_clk_gen;	// module which is in charge of producing this signal

  lost_val_alerter<T_WRITE>* alerter;	// module which checks for SC-TDF undersampling

  TUV_ams_lib_signal<MOC_WRITE,T_WRITE>* insignal;	// the input signal

  sca_eln::sca_node* internal_elec_node_pos; 	// the internal electrical nodes+-
  sca_eln::sca_node* internal_elec_node_neg;

  sc_signal<T_WRITE>* outsignal_sc_0;	// pointers for sc_signals to be passed to reading ports
  sc_signal<T_READ1>* outsignal_sc_1;  	// the T_WRITE-signal outsignal_sc_0 will only be created
  sc_signal<T_READ2>* outsignal_sc_2;  	// if the writing port is _not_ an sc-Port

  sca_tdf::sca_signal<T_WRITE>* outsignal_tdf_0;	// The same as above for tdf
  sca_tdf::sca_signal<T_READ1>* outsignal_tdf_1;
  sca_tdf::sca_signal<T_READ2>* outsignal_tdf_2;

  sc_fifo<T_WRITE>* outsignal_fifo_0;		// The same as above for sc_fifos
  sc_fifo<T_READ1>* outsignal_fifo_1;
  sc_fifo<T_READ2>* outsignal_fifo_2;

  converter<MOC_WRITE,SC,T_WRITE,T_WRITE>* conv_2sc_0;   	// pointer to the respective converters
  converter<MOC_WRITE,SC,T_WRITE,T_READ1>* conv_2sc_1;	   	// they will be instantiated only when
  converter<MOC_WRITE,SC,T_WRITE,T_READ2>* conv_2sc_2;		// needed

  converter<MOC_WRITE,TDF,T_WRITE,T_WRITE>* conv_2tdf_0;
  converter<MOC_WRITE,TDF,T_WRITE,T_READ1>* conv_2tdf_1;
  converter<MOC_WRITE,TDF,T_WRITE,T_READ2>* conv_2tdf_2;

  converter<MOC_WRITE,FIFO,T_WRITE,T_WRITE>* conv_2fifo_0;
  converter<MOC_WRITE,FIFO,T_WRITE,T_READ1>* conv_2fifo_1;
  converter<MOC_WRITE,FIFO,T_WRITE,T_READ2>* conv_2fifo_2;

  converter<MOC_WRITE,ELEC_VOLTAGE,T_WRITE,double>* conv_2voltage;  // the converters to ELEC_X always
  								    // have reading data type double,
  converter<MOC_WRITE,ELEC_CURRENT,T_WRITE,double>* conv_2current;  // indpendand of T_READ!, T_READ2

  bool writer_connected;	// is true if we have an input signal, false otherwise
  MoC elec_mode; // determines if from/to electrical network conversion uses voltage or current
  double elec_scaling_factor;	// stores factor for gain of to/from ELEC_X comversion

  int readingFIFOSize;	// the size of the internal fifo on the reading side

  T_WRITE const_val_empty_fifo_0;  // variables holding values to be used in case
  T_READ1 const_val_empty_fifo_1;  // of an empty reading side internal fifo
  T_READ2 const_val_empty_fifo_2;

  converterchannel_options<T_WRITE, T_WRITE>* options_0;	// options to be passed to the converters
  converterchannel_options<T_WRITE, T_READ1>* options_1;
  converterchannel_options<T_WRITE, T_READ2>* options_2;
  converterchannel_options<T_WRITE, double>* options_D;

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class constructor ////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

  void initialize()
  {
    internal_elec_node_pos = 0;
    internal_elec_node_neg = 0;
    outsignal_sc_0 = 0;
    outsignal_sc_1 = 0;
    outsignal_sc_2 = 0;
    outsignal_tdf_0 = 0;
    outsignal_tdf_1 = 0;
    outsignal_tdf_2 = 0;
    outsignal_fifo_0 = 0;
    outsignal_fifo_1 = 0;
    outsignal_fifo_2 = 0;

    tdf_clk = 0;

    elec_mode = NONE;
    readingFIFOSize=16+1;	// +1 for TDF/SC->FIFO conversion issues

    const_val_empty_fifo_0 = (T_WRITE)0;
    const_val_empty_fifo_1 = (T_READ1)0;
    const_val_empty_fifo_2 = (T_READ2)0;

    options_0 = new converterchannel_options<T_WRITE, T_WRITE>;
    options_1 = new converterchannel_options<T_WRITE, T_READ1>;
    options_2 = new converterchannel_options<T_WRITE, T_READ2>;
    options_D = new converterchannel_options<T_WRITE, double>;  // extra for conversion to ELEC_X

    writer_connected = false;
  }

  #define cc_delete_set0(X) delete X; X=0;
  public:
  ~converterchannel() {
    cc_delete_set0(options_0);
    cc_delete_set0(options_1);
    cc_delete_set0(options_2);
    cc_delete_set0(options_D);
  }

  public:

  explicit
  converterchannel(sc_module_name nm=sc_gen_unique_name("converterchannel")){
    if (MOC_WRITE == FIFO)
    {
      ostringstream warn_str;
      warn_str << "No internal buffer size of converter channel " << name() << endl;
      warn_str << "specified. Using default value of 16!" << endl;
      SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str.str().c_str());
      insignal = new TUV_ams_lib_signal<MOC_WRITE,T_WRITE>(16);
    } else
      insignal = new TUV_ams_lib_signal<MOC_WRITE,T_WRITE>;
    initialize();
  }

  explicit
  converterchannel(int n, sc_module_name nm=sc_gen_unique_name("converterchannel"))
    : sc_channel( nm )
  {
    if (MOC_WRITE != FIFO)
    {
      ostringstream warn_str;
      warn_str << "ERROR: integer parameter in converter channel " << name() << endl;
      warn_str << "       constructor only valid if writing MoC is set to FIFO!" << endl;
      SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
    }
    else
    {
      insignal = new TUV_ams_lib_signal<MOC_WRITE,T_WRITE>(n);
      initialize();
    }
  }

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// class methods & operators ////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// options //////////////////////////////////////////////////////////////////////////////////////////////////

  void float2IntPolicy(Float2IntPol pol)
  {
    options_1->f2ipol=pol;
    options_2->f2ipol=pol;
    options_D->f2ipol=pol;
  }

  void overflowClipping()
  {
    options_1->oflowpol=CLIP;
    options_2->oflowpol=CLIP;
    options_D->oflowpol=CLIP;
  }

private:


// TODO
// Was ist wenn T_WRITE unsigned ???

public:

  template<typename T_READ_X>
  void setRangeScaling(T_WRITE min, T_WRITE max)
  {
    converterchannel_options<T_WRITE,T_READ_X> *options=0;

    if  (typeid(T_READ_X)==typeid(T_READ1))
      options=reinterpret_cast<converterchannel_options<T_WRITE,T_READ_X> *> (options_1);
    else if (typeid(T_READ_X)==typeid(T_READ2))
       options=reinterpret_cast<converterchannel_options<T_WRITE,T_READ_X> *> (options_2);

    if (! options) {
      ostringstream warn_str;
      warn_str << "ERROR: setRangeScaling called for invalid read type, in converterchannel:" << name() << endl;
      SC_REPORT_ERROR("AHES-Converterchannels",warn_str.str().c_str());
      return;
    }

    options->setRangeScaling(min,max);
    CONVERTERS_TRACE_CONV_DO(std::cerr << "converterchannel " << name() << " : value range scaling for read type "<< demangle<T_READ_X>() <<" set to " << min << " - " << max << endl;)
/*    options_1->setRangeScaling(min, max);
    options_2->setRangeScaling(min, max);*/
  }

  void setFIFO2tdfemptybuffer(tdf_FIFO_OPT opt)
  {
    options_0->FIFO2tdfemptybuffer = opt;
    options_1->FIFO2tdfemptybuffer = opt;
    options_2->FIFO2tdfemptybuffer = opt;
  }

  void settdf2FIFOfullbuffer(tdf_FIFO_OPT opt)
  {
    options_0->tdf2FIFOfullbuffer = opt;
    options_1->tdf2FIFOfullbuffer = opt;
    options_2->tdf2FIFOfullbuffer = opt;
  }

  void setSC2FIFOfullbuffer(tdf_FIFO_OPT opt)
  {
    options_0->SC2FIFOfullbuffer = opt;
    options_1->SC2FIFOfullbuffer = opt;
    options_2->SC2FIFOfullbuffer = opt;
  }

  void setELEC2FIFOfullbuffer(tdf_FIFO_OPT opt)
  {
    options_0->ELEC2FIFOfullbuffer = opt;
    options_1->ELEC2FIFOfullbuffer = opt;
    options_2->ELEC2FIFOfullbuffer = opt;

    options_0->SC2FIFOfullbuffer = opt;		// actually, these setting here are the important ones, since
    options_1->SC2FIFOfullbuffer = opt;		// internally, and SC->FIFO converter is used. The ELEC2FIFOfullbuffer
    options_2->SC2FIFOfullbuffer = opt;    	// is kept for possible different implementations in the future
  }

  void setReadingFIFOSize(int n)
  {
    bool used=(outsignal_fifo_0!=0)||(outsignal_fifo_1!=0)||(outsignal_fifo_2!=0);
    if (used)
    {
      ostringstream error_str;
      error_str << "ERROR: converter channel already connected to a FIFO reader!" << endl;
      error_str << "       please call the method setReadingFIFOSize(int) " << endl;
      error_str << "       _before_ connecting to a FIFO reader" << endl;
      SC_REPORT_ERROR("ANDRES-Converterchannels",error_str.str().c_str());
    } else
    readingFIFOSize = n+1;	// +1 for tdf/SC->FIFO conversion issues
  }

  void setFIFO2tdfemptybufferConstant(T_WRITE val)
  {
     options_0->const_val_empty_fifo = val;
  }

  void setFIFO2tdfemptybufferConstant(T_READ1 val)
  {
     options_1->const_val_empty_fifo = val;
  }

  void setFIFO2tdfemptybufferConstant(T_READ2 val)
  {
     options_2->const_val_empty_fifo = val;
  }

  void use_lost_val_alerter()
  {
    options_0->use_lost_val_alerter = true;
    options_1->use_lost_val_alerter = true;
    options_2->use_lost_val_alerter = true;
  }

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// providing the clock signal which ticks with the speed of the tdf-cluster /////////////////////////////////

  sc_signal<bool>& clock_tdf()
  {
    if(tdf_clk==0) {

      tdf_clk= new sc_signal<bool>;
      tdf_clk_gen = new tdf_clock( "tdf_clock" );
      tdf_clk_gen->out( *tdf_clk );

      options_0->tdf_clk_gen = tdf_clk_gen;
      options_1->tdf_clk_gen = tdf_clk_gen;
      options_2->tdf_clk_gen = tdf_clk_gen;
      options_D->tdf_clk_gen = tdf_clk_gen;
    }
    return *tdf_clk;
  }

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// connecting to sc-ports ///////////////////////////////////////////////////////////////////////////////////

  operator sc_signal<T_WRITE>& ()
  {
    if(MOC_WRITE!=SC)
    {
      if(outsignal_sc_0==0){					// we have no reader yet?
        outsignal_sc_0 = new sc_signal<T_WRITE>;	// -> create the signal
      }
      return *outsignal_sc_0;
    }
    else
    {
      writer_connected = true;
      return *(dynamic_cast<sc_signal<T_WRITE>*>(insignal));
    }
  }

  operator sc_signal<T_READ1>& ()
  {						// these operators get called if a reading sc-port of datatype
    if(outsignal_sc_1==0){  				// T_READ1-T_READ5 connects to the converterchannel.
      outsignal_sc_1 = new sc_signal<T_READ1>;	// and create an approriate sc_signal, which is then returned
    }   					// such that _this_ signal is actually connected to the port.
    return *outsignal_sc_1;			// In the before_end_of_elaboration-Method, this signal will then
  }						// be connected to an appropriate converter.

  operator sc_signal<T_READ2>& ()
  {
    if(outsignal_sc_2==0){				// signal already created?
      outsignal_sc_2 = new sc_signal<T_READ2>;	// if not, do so
    }
    return *outsignal_sc_2;			// return the signal
  }


/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// connecting to tdf-ports on the reading side //////////////////////////////////////////////////////////////

  operator sca_tdf::sca_signal<T_WRITE>& ()
  {
    if(MOC_WRITE!=TDF)
    {
      if(outsignal_tdf_0==0){
        outsignal_tdf_0 = new sca_tdf::sca_signal<T_WRITE>;
      }
      return *outsignal_tdf_0;
    }
    else
    {
      writer_connected = true;
      return *(dynamic_cast<sca_tdf::sca_signal<T_WRITE>*>(insignal));
    }
  }

  operator sca_tdf::sca_signal<T_READ1>& ()
  {
    if(outsignal_tdf_1==0){
      outsignal_tdf_1 = new sca_tdf::sca_signal<T_READ1>;
    }
    return *outsignal_tdf_1;
  }

  operator sca_tdf::sca_signal<T_READ2>& ()
  {
    if(outsignal_tdf_2==0){
      outsignal_tdf_2 = new sca_tdf::sca_signal<T_READ2>;
    }
    return *outsignal_tdf_2;
  }

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// connecting to fifo-ports on the reading side////////////////////////////////////////////////////////////////

  operator sc_fifo<T_WRITE>& ()
  {
    if(MOC_WRITE!=FIFO)
    {
      if(outsignal_fifo_0==0){
        outsignal_fifo_0 = new sc_fifo<T_WRITE>(readingFIFOSize);
      }
      return *outsignal_fifo_0;
    }
    else
    {
      writer_connected = true;
      return *(dynamic_cast<sc_fifo<T_WRITE>*>(insignal));
    }
  }

  operator sc_fifo<T_READ1>& ()
  {
    if(outsignal_fifo_1==0){
      outsignal_fifo_1 = new sc_fifo<T_READ1>(readingFIFOSize);
    }
    return *outsignal_fifo_1;
  }

  operator sc_fifo<T_READ2>& ()
  {
    if(outsignal_fifo_2==0){
      outsignal_fifo_2 = new sc_fifo<T_READ2>(readingFIFOSize);
    }
    return *outsignal_fifo_2;
  }

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// connecting to electrical ports ///////////////////////////////////////////////////////////////////////////

  void setElec_Mode(MoC _elec_mode)
  {
    elec_mode = _elec_mode;
  }

  void setElec_Scaling(double val)
  {
    options_0->elec_scaling_factor = val;
    options_1->elec_scaling_factor = val;
    options_2->elec_scaling_factor = val;
    options_D->elec_scaling_factor = val;
  }

  void setTerminating_Resistance(double val)
  {
    if(MOC_WRITE!=ELEC_CURRENT)
    {
      ostringstream warn_str;
      warn_str << "converter channel " << name() << ":" << endl;
      warn_str << "It is not allowed to set a terminating resistance for a" << endl;
      warn_str << "writing MoC other than ahes::ELEC_CURRENT!" << endl;
      SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
    }
    options_0->terminating_resistance = val;
    options_1->terminating_resistance = val;
    options_2->terminating_resistance = val;
    // makes no sense for options_D !
  }

  operator sca_eln::sca_node& ()
  {
    if(internal_elec_node_pos==0)
    {
      if(MOC_WRITE==ELEC_VOLTAGE || MOC_WRITE==ELEC_CURRENT)
      {
        writer_connected = true;
        internal_elec_node_pos = dynamic_cast<sca_eln::sca_node*>(insignal);
      }
      else internal_elec_node_pos = new sca_eln::sca_node;
    }
    return *internal_elec_node_pos;
  }

  sca_eln::sca_node& neg_elec_node()
  {
    if(internal_elec_node_neg==0)
    {
      internal_elec_node_neg = new sca_eln::sca_node;
      options_0->elec_node_neg = internal_elec_node_neg;
      options_1->elec_node_neg = internal_elec_node_neg;
      options_2->elec_node_neg = internal_elec_node_neg;
      options_D->elec_node_neg = internal_elec_node_neg;
    }
    return *internal_elec_node_neg;
  }

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// instantiating and connecting converters where needed /////////////////////////////////////////////////////
  private:

  void before_end_of_elaboration()  	// this is done just before simulation starts
  {
    if(!writer_connected)
    {
      ostringstream warn_str;
      warn_str << "ERROR: No writing port connected for converter channel " << name() << endl;
      SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
    }
    else
    {
      if (MOC_WRITE==FIFO) {
        uint outcount=0;
        if (outsignal_fifo_0) outcount++;
        if (outsignal_fifo_1) outcount++;
        if (outsignal_fifo_2) outcount++;
        if (outsignal_tdf_0) outcount++;
        if (outsignal_tdf_1) outcount++;
        if (outsignal_tdf_2) outcount++;
        if (outcount>1) {
          ostringstream warn_str;
          warn_str << "ERROR: Writer is FIFO, so only one reader can be connected to converter channel " << name() << endl;
          SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
        }
      }
      // SC reading
      if(outsignal_sc_0!=0 && MOC_WRITE != SC)
      {
        conv_2sc_0 = new converter<MOC_WRITE,SC,T_WRITE, T_WRITE>("conv_2sc_0", options_0);
        conv_2sc_0->in(*insignal);
        conv_2sc_0->out(*outsignal_sc_0);
      }
      if(outsignal_sc_1!=0 )
      {
        conv_2sc_1 = new converter<MOC_WRITE,SC,T_WRITE, T_READ1>("conv_2sc_1", options_1);
        conv_2sc_1->in(*insignal);
        conv_2sc_1->out(*outsignal_sc_1);
      }
      if(outsignal_sc_2!=0 )
      {
        conv_2sc_2 = new converter<MOC_WRITE,SC,T_WRITE, T_READ2>("conv_2sc_2", options_2);
        conv_2sc_2->in(*insignal);
        conv_2sc_2->out(*outsignal_sc_2);
      }
      // tdf reading
      if(outsignal_tdf_0!=0  && MOC_WRITE != TDF)
      {
        CONVERTERS_TRACE_CONV_DO(std::cerr << "TDF 0 reading" << endl;)
	conv_2tdf_0 = new converter<MOC_WRITE, TDF, T_WRITE, T_WRITE>("conv_2tdf_0", options_0);
        conv_2tdf_0->in(*insignal);
        conv_2tdf_0->out(*outsignal_tdf_0);
      }
      if(outsignal_tdf_1!=0)
      {
	conv_2tdf_1 = new converter<MOC_WRITE, TDF, T_WRITE, T_READ1>("conv_2tdf_1", options_1);
        conv_2tdf_1->in(*insignal);
        conv_2tdf_1->out(*outsignal_tdf_1);
      }
      if(outsignal_tdf_2!=0)
      {
	conv_2tdf_2 = new converter<MOC_WRITE, TDF, T_WRITE, T_READ2>("conv_2tdf_2", options_2);
        conv_2tdf_2->in(*insignal);
        conv_2tdf_2->out(*outsignal_tdf_2);
      }
      // FIFO reding
      if(outsignal_fifo_0!=0 && MOC_WRITE != FIFO)
      {
        conv_2fifo_0 = new converter<MOC_WRITE, FIFO, T_WRITE, T_WRITE>("conv_2fifo_0", options_0);
        conv_2fifo_0->in(*insignal);
        conv_2fifo_0->out(*outsignal_fifo_0);
      }
      if(outsignal_fifo_1!=0)
      {
        conv_2fifo_1 = new converter<MOC_WRITE, FIFO, T_WRITE, T_READ1>("conv_2fifo_1", options_1);
        conv_2fifo_1->in(*insignal);
        conv_2fifo_1->out(*outsignal_fifo_1);
      }
      if(outsignal_fifo_2!=0)
      {
        conv_2fifo_2 = new converter<MOC_WRITE, FIFO, T_WRITE, T_READ2>("conv_2fifo_2", options_2);
        conv_2fifo_2->in(*insignal);
        conv_2fifo_2->out(*outsignal_fifo_2);
      }


      // ELEC_CURRENT or ELEC_VOLTAGE reading
      if(internal_elec_node_pos!=0 && MOC_WRITE!=ELEC_VOLTAGE && MOC_WRITE!=ELEC_CURRENT)
      {
        if(elec_mode==NONE)
	{
    	  ostringstream warn_str;
    	  warn_str << "converter channel " << name() << ":" << endl;
    	  warn_str << "No electrical reading mode (voltage or current) for ELEC reading side set!" << endl;
    	  warn_str << "use the converter channel method setElec_Mode(mode) with " << endl;
    	  warn_str << "mode = ahes::ELEC_CURRENT or ahes::ELEC_VOLTAGE!" << endl;
    	  SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
	}
        if(elec_mode==ELEC_VOLTAGE)
	{
          conv_2voltage = new converter<MOC_WRITE, ELEC_VOLTAGE, T_WRITE, double>("conv_2voltage", options_D);
          conv_2voltage->in(*insignal);
          conv_2voltage->out(*internal_elec_node_pos);
	}
        if(elec_mode==ELEC_CURRENT)
	{
          conv_2current = new converter<MOC_WRITE, ELEC_CURRENT, T_WRITE, double>("conv_2current", options_D);
          conv_2current->in(*insignal);
          conv_2current->out(*internal_elec_node_pos);
	}
      }
    }
  }

  void end_of_simulation()
  {
  }

};


} // end namespace ahes

#endif // AHES_CONVERTER_CONVERTERCHANNEL_H
// kate: space-indent on; indent-width 2; replace-tabs on;
