//-------------------------------------------------------------------------------------------
// Copyright (c) 2008-2010, Institute of Computer Technology, University of Technology Vienna
// Authors:
//   Christoph Grimm, Jan Haase, Markus Damm, Jiong Ou,  
//   Joseph Wenninger, Peter Brunmayr, Muhammad Farooq
// Contact:
//   tuv-ams-lib@ict.tuwien.ac.at
// For the license please read LICENSE.txt
//-------------------------------------------------------------------------------------------

#ifndef AHES_CONVERTER_CONVERTERS_H
#define AHES_CONVERTER_CONVERTERS_H 1

#include "utils/TUV_ams_lib_moc.h"
#include "converters_data.h"

namespace TUV_ams_lib {

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// data type conversion class needed by some MoC converters /////////////////////////////////////////////////

template <typename T_WRITE, typename T_READ>
class datconv_tdf: sca_tdf::sca_module, public datconv<T_WRITE, T_READ>{

public:

  sca_tdf::sca_in<T_WRITE> in;
  sca_tdf::sca_out<T_READ> out;

  datconv_tdf(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options) : datconv<T_WRITE, T_READ>(options){}

  void sig_proc()
  {
    T_WRITE val = in.read();
    out.write(conv(val));
  }
};

template <typename T_WRITE, typename T_READ>
class datconv_sc: sc_module, public datconv<T_WRITE, T_READ>{

public:

  sc_in<T_WRITE> in;
  sc_out<T_READ> out;

  SC_HAS_PROCESS(datconv_sc);
  datconv_sc(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options) : datconv<T_WRITE, T_READ>(options)
  {
    SC_METHOD(do_it);
    sensitive << in;
  }

  void do_it()
  {
    T_WRITE val = in.read();
    out.write(conv(val));
  }
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// converter modules for conversion between MoCs ////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

// The following modules provide the conversion between different MoCs, e.g. from SC to tdf, but also
// between different datatypes using the same MoC. To this end, they inherit from the appropriately
// templated converter_base class

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 0 general templated class ********************************************************************************
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

template <MoC M_WRITE, MoC M_READ, typename T_WRITE, typename T_READ>
class converter: sc_module{

  public:
  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options) : datconv<T_WRITE, T_READ>(options){}

};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 1 conversion from SC (DE) ********************************************************************************
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion SC->SC ////////////////////////////////////////////////////////////////////////////////////////

template <typename T_WRITE, typename T_READ>
class converter<SC, SC, T_WRITE, T_READ>: sc_module, public datconv<T_WRITE, T_READ>{	// inherit from the converter-base class
								// which provides the conv-Methods to
  public:							// convert to T_READ
  sc_in<T_WRITE> in;    					// the next modules work the same
  sc_out<T_READ> out;						// but use appropriate converter-ports
								// for MoC-conversion when needed

  SC_HAS_PROCESS(converter);
  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options) : datconv<T_WRITE, T_READ>(options)
  {
    SC_METHOD(do_it);
    sensitive << in;
  }

  private:
  void do_it() {
    out.write(conv(in.read()));
  };
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion SC->tdf ///////////////////////////////////////////////////////////////////////////////////////

template <typename T_WRITE>
SC_MODULE(lost_val_alerter) {

  public:
  sc_in<T_WRITE> in;

  int valcounter;

  SC_CTOR(lost_val_alerter) {
    SC_METHOD(do_it);
    sensitive << in;
    valcounter=0;
  }

  private:
  void do_it() {
    valcounter++;
  };

  public:

  int get_count()
  {
    int zw = valcounter-1;
    if(zw<0) zw=0;
    valcounter = 0;
    return zw;
  }
};

template <typename T_WRITE, typename T_READ>
class converter_sc2tdf_sub: sca_tdf::sca_module, public datconv<T_WRITE, T_READ>{

  public:
  sca_tdf::sca_de::sca_in<T_WRITE> in;	// converter port SC <-> tdf
  sca_tdf::sca_out<T_READ> out;

  bool using_alerter;

  lost_val_alerter<T_WRITE>* alerter;

  converter_sc2tdf_sub(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options) : datconv<T_WRITE, T_READ>(options)
  {
    using_alerter = false;
  }

  converter_sc2tdf_sub(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options, lost_val_alerter<T_WRITE>* al) : datconv<T_WRITE, T_READ>(options)
  {
    alerter=al;
    using_alerter = true;
  }

  private:
  void processing() {
    out.write(conv(in.read()));
    if(using_alerter)
    {
      int lost_vals = alerter->get_count();
      if(lost_vals>0)
      {
        ostringstream warn_str;
        warn_str << "Sample period of tdf-cluster with converter channel " << get_parent()->name() << endl;
        warn_str << "to long; lost " << lost_vals << " values from DE side at " << get_time() << "!" << endl;
        SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str.str().c_str());
      }
    }
  };
};

template <typename T_WRITE, typename T_READ>
class converter<SC, TDF, T_WRITE, T_READ>: sc_module{

  public:
  sc_in<T_WRITE> in;	// converter port SC <-> tdf
  sca_tdf::sca_out<T_READ> out;

  lost_val_alerter<T_WRITE>* alerter;

  converter_sc2tdf_sub<T_WRITE, T_READ>* tdf_sub;

  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options)
  {
    if(options->use_lost_val_alerter)
    {
      alerter = new lost_val_alerter<T_WRITE>("alerter");
      alerter->in(in);
      tdf_sub = new converter_sc2tdf_sub<T_WRITE, T_READ>("tdf_sub",options,alerter);
    }
    else tdf_sub = new converter_sc2tdf_sub<T_WRITE, T_READ>("tdf_sub", options);
    tdf_sub->in(in);
    tdf_sub->out(out);
  }

};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion SC->FIFO  /////////////////////////////////////////////////////////////////////////////////////

template <typename T_WRITE, typename T_READ>
class converter<SC, FIFO, T_WRITE, T_READ>: sc_module, public datconv<T_WRITE, T_READ>{

  private:
  tdf_FIFO_OPT SC2FIFOfullbuffer;
  sc_fifo<T_READ>* con_fifo;

  public:
  sc_in<T_WRITE> in;
  sc_fifo_out<T_READ> out;

  SC_HAS_PROCESS(converter);
  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options) : datconv<T_WRITE, T_READ>(options)
  {
    SC_METHOD(do_it);
    sensitive << in;
    dont_initialize();
    SC2FIFOfullbuffer = options->SC2FIFOfullbuffer;
  }

  private:

  void before_end_of_elaboration()
  {
    con_fifo = dynamic_cast<sc_fifo<T_READ>*>(out.get_interface());
  }

  void do_it()
  {
    if (con_fifo==0) con_fifo = dynamic_cast<sc_fifo<T_READ>*>(out.get_interface());
    T_READ val=conv(in.read());
    T_READ discVAl;
    if(con_fifo->num_free()>1)	// write possible?
      out.nb_write(val);
    else
    {
      ostringstream warn_str;
      switch(SC2FIFOfullbuffer)
      {
        case error:
          warn_str << "internal buffer of converter channel " << get_parent()->name() << " full!" << endl;
          warn_str << "       change the handling of this case by using the converter channel method" << endl;
          warn_str << "       setSC2FIFOfullbuffer(option) where option is" << endl;
          warn_str << "       ahes::error    : this very behaviour" << endl;
          warn_str << "       ahes::discardOldest  : the oldest value of the internal buffer" << endl;
          warn_str << "                              will be discarded." << endl;
          warn_str << "       ahes::discardCurrent : the value which should be written to the" << endl;
          warn_str << "                              inernal buffer will be discarded." << endl;
          SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
        case discardOldest:
	  con_fifo->nb_read(discVAl);
          warn_str << "internal buffer of converter channel " << get_parent()->name() << " full!" << endl;
	  warn_str << "discarding oldest value (" << discVAl << ") in buffer" << endl;
          SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str.str().c_str());
	  break;
        case discardCurrent:
          warn_str << "internal buffer of converter channel " << get_parent()->name() << " full!" << endl;
	  warn_str << "discarding the value to be written" << endl;
          SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str.str().c_str());
	  break;
	default: sc_assert(false);
      }
    }
  };
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion SC->ELEC_VOLTAGE //////////////////////////////////////////////////////////////////////////////

template <typename T_WRITE, typename T_READ>	// Note: T_READ will always be double
class converter<SC, ELEC_VOLTAGE, T_WRITE, T_READ>: sc_module{

  sca_eln::sca_de_vsource* vconv;
  sca_eln::sca_node_ref gnd;

  datconv_sc<T_WRITE, double>* dconv;

  sc_signal<double> sig;

  public:

  sc_in<T_WRITE> in;
  sca_eln::sca_terminal out;

  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options)
  {
    vconv = new sca_eln::sca_de_vsource("vconv");
      vconv->scale = options->elec_scaling_factor;
      vconv->inp(sig);
      vconv->p(out);
      if(options->elec_node_neg!=0) vconv->n(*(options->elec_node_neg));
      else vconv->n(gnd);

    dconv = new datconv_sc<T_WRITE, double>("dconv", options);
      dconv->in(in);
      dconv->out(sig);
  }
};

template <>	// specialization for T_WRITE = double => no datatype conversion needed
class converter<SC, ELEC_VOLTAGE, double, double>: sc_module{

  sca_eln::sca_de_vsource* vconv;
  sca_eln::sca_node_ref gnd;

  public:

  sc_in<double> in;
  sca_eln::sca_terminal out;

  converter(sc_core::sc_module_name n, converterchannel_options<double, double>* options)
  {
    vconv = new sca_eln::sca_de_vsource("vconv");
      vconv->scale = options->elec_scaling_factor;
      vconv->inp(in);
      vconv->p(out);
      if(options->elec_node_neg!=0) vconv->n(*(options->elec_node_neg));
      else vconv->n(gnd);
  }
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion SC->ELEC_CURRENT //////////////////////////////////////////////////////////////////////////////

template <typename T_WRITE, typename T_READ>	// Note: T_READ will always be double
class converter<SC, ELEC_CURRENT, T_WRITE, T_READ>: sc_module{

  sca_eln::sca_de_isource* iconv;
  sca_eln::sca_node_ref gnd;

  datconv_sc<T_WRITE, double>* dconv;

  sc_signal<double> sig;

  public:

  sc_in<T_WRITE> in;
  sca_eln::sca_terminal out;

  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options)
  {
    iconv = new sca_eln::sca_de_isource("iconv");
      iconv->scale = options->elec_scaling_factor;
      iconv->inp(sig);
      iconv->p(out);
      if(options->elec_node_neg!=0) iconv->n(*(options->elec_node_neg));
      else iconv->n(gnd);

    dconv = new datconv_sc<T_WRITE, double>("dconv", options);
      dconv->in(in);
      dconv->out(sig);
  }
};

template <>	// specialization for T_WRITE = double => no datatype conversion needed
class converter<SC, ELEC_CURRENT, double, double>: sc_module{

 sca_eln::sca_de_isource* iconv;
  sca_eln::sca_node_ref gnd;

  public:

  sc_in<double> in;
  sca_eln::sca_terminal out;

  converter(sc_core::sc_module_name n, converterchannel_options<double, double>* options)
  {
    iconv = new sca_eln::sca_de_isource("iconv");
      iconv->scale = options->elec_scaling_factor;
      iconv->inp(in);
      iconv->p(out);
      if(options->elec_node_neg!=0) iconv->n(*(options->elec_node_neg));
      else iconv->n(gnd);
  }
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 2 conversion from tdf ************************************************************************************
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion tdf->SC ///////////////////////////////////////////////////////////////////////////////////////

template <typename T_WRITE, typename T_READ>
class converter<TDF, SC, T_WRITE, T_READ>: sca_tdf::sca_module, public datconv<T_WRITE, T_READ>{

  public:
  sca_tdf::sca_in<T_WRITE> in;
  sca_tdf::sca_de::sca_out<T_READ> out;	// converter port SC <-> tdf

  tdf_clock* tdf_clk_gen;

  bool clock_started, using_tdf_clock;

  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options) : datconv<T_WRITE, T_READ>(options)
  {
    if(options->tdf_clk_gen!=0)
    {
      tdf_clk_gen=options->tdf_clk_gen;
      using_tdf_clock = true;
      clock_started=false;
    }
    else
    {
      using_tdf_clock = false;
      clock_started=true;  // programming trick to make evaluating the if clause
                         // within the sig_proc faster
    }
  }

  private:
  void sig_proc() {
    out.write(conv(in.read()));
    if(!clock_started && using_tdf_clock)
    {
      clock_started = true;
      tdf_clk_gen->start(in.get_T());
    }
  };
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion tdf->tdf //////////////////////////////////////////////////////////////////////////////////////

template <typename T_WRITE, typename T_READ>
class converter<TDF, TDF, T_WRITE, T_READ>: sca_tdf::sca_module, public datconv<T_WRITE, T_READ>{

  public:
  sca_tdf::sca_in<T_WRITE> in;
  sca_tdf::sca_out<T_READ> out;

  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options) : datconv<T_WRITE, T_READ>(options){}

  private:
  void sig_proc() {
    out.write(conv(in.read()));
  };
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion tdf->FIFO /////////////////////////////////////////////////////////////////////////////////////

template <typename T_WRITE, typename T_READ>
class converter<TDF, FIFO, T_WRITE, T_READ>: sca_tdf::sca_module, public datconv<T_WRITE, T_READ>{

  private:
  int warningCount;
  tdf_FIFO_OPT tdf2FIFOfullbuffer;
  sc_fifo<T_READ>* con_fifo;

  public:
  sca_tdf::sca_in<T_WRITE> in;
  sc_fifo_out<T_READ> out;

  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options) : datconv<T_WRITE, T_READ>(options)
  {
    tdf2FIFOfullbuffer = options->tdf2FIFOfullbuffer;
  }

  private:

  void init()
  {
    con_fifo = dynamic_cast<sc_fifo<T_READ>*>(out.get_interface());
  }

  void sig_proc()
  {
    if (con_fifo==0) con_fifo = dynamic_cast<sc_fifo<T_READ>*>(out.get_interface());
    T_READ discVAl;
    T_READ val=conv(in.read());
    if(con_fifo->num_free()>1)	// enough space?
      out.nb_write(val);
    else
    {
      ostringstream warn_str;
      switch(tdf2FIFOfullbuffer)
      {
        case error:
          warn_str << "internal buffer of converter channel " << get_parent()->name() << " full!" << endl;
          warn_str << "       change the handling of this case by using the converter channel method" << endl;
          warn_str << "       settdf2FIFOfullbuffer(option) where option is" << endl;
          warn_str << "       ahes::error    : this very behaviour" << endl;
          warn_str << "       ahes::discardOldest  : the oldest value of the internal buffer" << endl;
          warn_str << "                              will be discarded." << endl;
          warn_str << "       ahes::discardCurrent : the value which should be written to the" << endl;
          warn_str << "                              inernal buffer will be discarded." << endl;
          SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
        case discardOldest:
	  con_fifo->nb_read(discVAl);
          warn_str << "internal buffer of converter channel " << get_parent()->name() << " full!" << endl;
	  warn_str << "discarding oldest value (" << discVAl << ") in buffer" << endl;
          SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str.str().c_str());
	  out.nb_write(val);
	  break;
        case discardCurrent:
          warn_str << "internal buffer of converter channel " << get_parent()->name() << " full!" << endl;
	  warn_str << "discarding the value to be written" << endl;
          SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str.str().c_str());
	  break;
	default: sc_assert(false);
      }
    }
  };
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion tdf->ELEC_VOLTAGE /////////////////////////////////////////////////////////////////////////////

template <typename T_WRITE, typename T_READ>	// Note: T_READ will always be double
class converter<TDF, ELEC_VOLTAGE, T_WRITE, T_READ>: sc_module{

  sca_eln::sca_tdf_vsource* vconv;
  sca_eln::sca_node_ref gnd;

  datconv_tdf<T_WRITE, double>* dconv;

  sca_tdf::sca_signal<double> sig;

  public:

  sca_tdf::sca_in<T_WRITE> in;
  sca_eln::sca_terminal out;

  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options)
  {
    vconv = new sca_eln::sca_tdf_vsource("vconv");
      vconv->scale = options->elec_scaling_factor;
      vconv->inp(sig);
      vconv->p(out);
      if(options->elec_node_neg!=0) vconv->n(*(options->elec_node_neg));
      else vconv->n(gnd);

    dconv = new datconv_tdf<T_WRITE, double>("dconv", options);
      dconv->in(in);
      dconv->out(sig);
  }
};

template <>	// specialization for T_WRITE = double => no datatype conversion needed
class converter<TDF, ELEC_VOLTAGE, double, double>: sc_module{

  sca_eln::sca_tdf_vsource* vconv;
  sca_eln::sca_node_ref gnd;

  public:

  sca_tdf::sca_in<double> in;
  sca_eln::sca_terminal out;

  converter(sc_core::sc_module_name n, converterchannel_options<double, double>* options)
  {
    vconv = new sca_eln::sca_tdf_vsource("vconv");
      vconv->scale = options->elec_scaling_factor;
      vconv->inp(in);
      vconv->p(out);
      if(options->elec_node_neg!=0) vconv->n(*(options->elec_node_neg));
      else vconv->n(gnd);
  }
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion tdf->ELEC_CURRENT /////////////////////////////////////////////////////////////////////////////

template <typename T_WRITE, typename T_READ>	// Note: T_READ will always be double
class converter<TDF, ELEC_CURRENT, T_WRITE, T_READ>: sc_module{

  sca_eln::sca_tdf_isource* iconv;
  sca_eln::sca_node_ref gnd;

  datconv_tdf<T_WRITE, double>* dconv;

  sca_tdf::sca_signal<double> sig;

  public:

  sca_tdf::sca_in<T_WRITE> in;
  sca_eln::sca_terminal out;

  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options)
  {
    iconv = new sca_eln::sca_tdf_isource("iconv");
      iconv->scale = options->elec_scaling_factor;
      iconv->inp(sig);
      iconv->p(out);
      if(options->elec_node_neg!=0) iconv->n(*(options->elec_node_neg));
      else iconv->n(gnd);

    dconv = new datconv_tdf<T_WRITE, double>("dconv",options);
      dconv->in(in);
      dconv->out(sig);
  }
};

template <>	// specialization for T_WRITE = double => no datatype conversion needed
class converter<TDF, ELEC_CURRENT, double, double>: sc_module{

  sca_eln::sca_tdf_isource* iconv;
  sca_eln::sca_node_ref gnd;

  public:

  sca_tdf::sca_in<double> in;
  sca_eln::sca_terminal out;

  converter(sc_core::sc_module_name n, converterchannel_options<double, double>* options)
  {
    iconv = new sca_eln::sca_tdf_isource("iconv");
      iconv->scale = options->elec_scaling_factor;
      iconv->inp(in);
      iconv->p(out);
      if(options->elec_node_neg!=0) iconv->n(*(options->elec_node_neg));
      else iconv->n(gnd);
  }
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 3 conversion from FIFO writers****************************************************************************
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion FIFO->SC  /////////////////////////////////////////////////////////////////////////////////////

template <typename T_WRITE, typename T_READ>
class converter<FIFO, SC, T_WRITE, T_READ>: sc_module{

  public:
  sc_fifo_in<T_WRITE> in;
  sc_out<T_READ> out;

  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options)
  {
    ostringstream warn_str;
    warn_str << "Converterchannel " << get_parent()->name() << ":" << endl;
    warn_str << "Conversion from FIFO to SC is not supported!" << endl;
    SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
  }
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion FIFO->TDF /////////////////////////////////////////////////////////////////////////////////////

template <typename T_WRITE, typename T_READ>
class converter<FIFO, TDF, T_WRITE, T_READ>: sca_tdf::sca_module, public datconv<T_WRITE, T_READ>{

  public:
  sc_fifo_in<T_WRITE> in;
  sca_tdf::sca_out<T_READ> out;
  sca_tdf::sca_de::sca_in<bool> sync;

  private:

  T_READ last_val;	// the last value that could be read successfully
  T_READ const_val;	// constant value when choosing empty buffer option constant

  tdf_FIFO_OPT FIFO2tdfemptybuffer; // option for handling empty buffers

  public:

  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options) : datconv<T_WRITE, T_READ>(options)
  {
    FIFO2tdfemptybuffer = options->FIFO2tdfemptybuffer;
    const_val =  options->const_val_empty_fifo;
	  sc_signal<bool>* sync_sig = new sc_signal<bool>;
	  this->sync(*sync_sig);
  }

  private:

  void sig_proc()
  {
    T_WRITE new_val;
    if(in.nb_read(new_val))	// read successful?
    {
      last_val=conv(new_val);	// => set last_val to new value
      out.write(last_val);
    }
    else if(get_time()==SC_ZERO_TIME)
    {
      out.write(const_val);
    }
    else
    {
      sync.read();
      if(in.nb_read(new_val))	// read successful?
      {
        last_val=conv(new_val);	// => set last_val to new value
        out.write(last_val);
      }
      else
      {
        ostringstream warn_str;
        switch(FIFO2tdfemptybuffer)
        {
          case error:
            warn_str << "internal buffer of converter channel " << get_parent()->name() << " empty!" << endl;
            warn_str << "       change the handling of this case by using the converter channel method" << endl;
            warn_str << "       setFIFO2tdfemptybuffer(option) where option is" << endl;
            warn_str << "       ahes::error    : this very behaviour" << endl;
            warn_str << "       ahes::hold     : the last succesful value read is returned" << endl;
            warn_str << "       ahes::constant : a constant value (writing side data type) is used which is" << endl;
            warn_str << "                        set with the method setFIFO2tdfemptybufferConstant(val)" << endl;
            SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
          case constant:
            warn_str << "internal buffer of converter channel " << get_parent()->name() << " empty, using constant value " << const_val << " instead" << endl;
            SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str.str().c_str());
	    out.write(const_val);
	    break;
          case hold:
            warn_str << "internal buffer of converter channel " << get_parent()->name() << " empty, using the last value " << last_val << " instead" << endl;
            SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str.str().c_str());
	    out.write(last_val);
	    break;
          default: sc_assert(false);
        }
      }
    }
  };
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion FIFO->FIFO ////////////////////////////////////////////////////////////////////////////////////

template <typename T_WRITE, typename T_READ>
class converter<FIFO, FIFO, T_WRITE, T_READ>: sc_module, public datconv<T_WRITE, T_READ>{

  public:
  sc_fifo_in<T_WRITE> in;
  sc_fifo_out<T_READ> out;

  SC_HAS_PROCESS(converter);
  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options) : datconv<T_WRITE, T_READ>(options)
  {
    SC_METHOD(do_it);
    sensitive << in;
  }

  private:
  void do_it() {
    out.write(conv(in.read()));
  };
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion FIFO->ELEC_VOLTAGE ////////////////////////////////////////////////////////////////////////////

template <typename T_WRITE, typename T_READ>	// Note: T_READ will always be double
class converter<FIFO, ELEC_VOLTAGE, T_WRITE, T_READ>: sc_module{

  public:

  sc_fifo_in<T_WRITE> in;
  sca_eln::sca_terminal out;

  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options)
  {
    ostringstream warn_str;
    warn_str << "Converterchannel " << get_parent()->name() << ":" << endl;
    warn_str << "Conversion from FIFO to ELEC_VOLTAGE is not supported!" << endl;
    SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
  }
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion FIFO->ELEC_CURRENT ////////////////////////////////////////////////////////////////////////////

template <typename T_WRITE, typename T_READ>	// Note: T_READ will always be double
class converter<FIFO, ELEC_CURRENT, T_WRITE, T_READ>: sc_module{

  public:

  sc_fifo_in<T_WRITE> in;
  sca_eln::sca_terminal out;

  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options)
  {
    ostringstream warn_str;
    warn_str << "Converterchannel " << get_parent()->name() << ":" << endl;
    warn_str << "Conversion from FIFO to ELEC_CURRENT is not supported!" << endl;
    SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
  }
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 4 conversion from ELEC_VOLTAGE ***************************************************************************
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion ELEC_VOLTAGE->SC //////////////////////////////////////////////////////////////////////////////

template <typename T_WRITE, typename T_READ>
class converter<ELEC_VOLTAGE, SC, T_WRITE, T_READ>: sc_module{

  public:

  sca_eln::sca_terminal in;
  sc_out<T_READ> out;

  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options)
  {
    ostringstream warn_str;
    warn_str << "converter channel " << get_parent()->name() << ":" << endl;
    warn_str << "When the writing side is an electrical network," << endl;
    warn_str << "the writing data type MUST be set to double!!!" << endl;
    SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
  }
};

template <typename T_READ>
class converter<ELEC_VOLTAGE, SC, double, T_READ>: sc_module{

  sca_eln::sca_de_vsink* vconv;
  datconv_sc<double, T_READ>* dconv;

  sc_signal<double> sig;
  sca_eln::sca_node_ref gnd;

  public:

  sca_eln::sca_terminal in;
  sc_out<T_READ> out;

  converter(sc_core::sc_module_name n, converterchannel_options<double, T_READ>* options)
  {
    vconv = new sca_eln::sca_de_vsink("vconv");
      vconv->scale = options->elec_scaling_factor;
      vconv->p(in);
      vconv->outp(sig);
      if(options->elec_node_neg!=0) vconv->n(*(options->elec_node_neg));
      else vconv->n(gnd);

    dconv = new datconv_sc<double, T_READ>("dconv", options);
      dconv->in(sig);
      dconv->out(out);
  }
};

template <>	// specialization for T_READ = double => no datatype conversion needed
class converter<ELEC_VOLTAGE, SC, double, double>: sc_module{

  sca_eln::sca_de_vsink* vconv;

  public:

  sca_eln::sca_terminal in;
  sc_out<double> out;
  sca_eln::sca_node_ref gnd;

  converter(sc_core::sc_module_name n, converterchannel_options<double, double>* options)
  {
    vconv = new sca_eln::sca_de_vsink("vconv");
      vconv->scale = options->elec_scaling_factor;
      vconv->p(in);
      vconv->outp(out);
      if(options->elec_node_neg!=0) vconv->n(*(options->elec_node_neg));
      else vconv->n(gnd);
  }
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion ELEC_VOLTAGE->tdf /////////////////////////////////////////////////////////////////////////////

template <typename T_WRITE, typename T_READ>
class converter<ELEC_VOLTAGE, TDF, T_WRITE, T_READ>: sc_module{

  public:

  sca_eln::sca_terminal in;
  sca_tdf::sca_out<T_READ> out;

  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options)
  {
    ostringstream warn_str;
    warn_str << "converter channel " << get_parent()->name() << ":" << endl;
    warn_str << "When the writing side is an electrical network," << endl;
    warn_str << "the writing data type MUST be set to double!!!" << endl;
    SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
  }
};

template <typename T_READ>
class converter<ELEC_VOLTAGE, TDF, double, T_READ>: sc_module{

  sca_eln::sca_tdf_vsink* vconv;
  datconv_tdf<double, T_READ>* dconv;

  sca_tdf::sca_signal<double> sig;
  sca_eln::sca_node_ref gnd;

  public:

  sca_eln::sca_terminal in;
  sca_tdf::sca_out<T_READ> out;

  converter(sc_core::sc_module_name n, converterchannel_options<double, T_READ>* options)
  {
    vconv = new sca_eln::sca_tdf_vsink("vconv");
      vconv->scale = options->elec_scaling_factor;
      vconv->p(in);
      vconv->outp(sig);
      if(options->elec_node_neg!=0) vconv->n(*(options->elec_node_neg));
      else vconv->n(gnd);

    dconv = new datconv_tdf<double, T_READ>("dconv", options);
      dconv->in(sig);
      dconv->out(out);
  }
};

template <>	// specialization for T_READ = double => no datatype conversion needed
class converter<ELEC_VOLTAGE, TDF, double, double>: sc_module{

  sca_eln::sca_tdf_vsink* vconv;

  public:

  sca_eln::sca_terminal in;
  sca_tdf::sca_out<double> out;
  sca_eln::sca_node_ref gnd;

  converter(sc_core::sc_module_name n, converterchannel_options<double, double>* options)
  {
    vconv = new sca_eln::sca_tdf_vsink("vconv");
      vconv->scale = options->elec_scaling_factor;
      vconv->p(in);
      vconv->outp(out);
      if(options->elec_node_neg!=0) vconv->n(*(options->elec_node_neg));
      else vconv->n(gnd);
  }
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion ELEC_VOLTAGE->FIFO ////////////////////////////////////////////////////////////////////////////

template <typename T_WRITE, typename T_READ>
class converter<ELEC_VOLTAGE, FIFO, T_WRITE, T_READ>: sc_module{

  public:

  sca_eln::sca_terminal in;
  sc_fifo_out<T_READ> out;

  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options)
  {
    ostringstream warn_str;
    warn_str << "converter channel " << get_parent()->name() << ":" << endl;
    warn_str << "When the writing side is an electrical network," << endl;
    warn_str << "the writing data type MUST be set to double!!!" << endl;
    SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
  }
};

template <typename T_READ>
class converter<ELEC_VOLTAGE, FIFO, double, T_READ>: sc_module{

  sca_eln::sca_de_vsink* vconv;
  converter<SC, FIFO, double, T_READ>* sc_fifo_conv;

  sc_signal<double> sig;
  sca_eln::sca_node_ref gnd;

  public:

  sca_eln::sca_terminal in;
  sc_fifo_out<T_READ> out;

  converter(sc_core::sc_module_name n, converterchannel_options<double, T_READ>* options)
  {
    vconv = new sca_eln::sca_de_vsink("vconv");
      vconv->scale = options->elec_scaling_factor;
      vconv->p(in);
      vconv->outp(sig);
      if(options->elec_node_neg!=0) vconv->n(*(options->elec_node_neg));
      else vconv->n(gnd);

    sc_fifo_conv = new converter<SC, FIFO, double, T_READ>("sc_fifo_conv", options);

      sc_fifo_conv->in(sig);
      sc_fifo_conv->out(out);
  }
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion ELEC_VOLTAGE->ELEC_CURRENT ////////////////////////////////////////////////////////////////////

// Will never be instantiated

template <typename T_WRITE, typename T_READ>
class converter<ELEC_VOLTAGE, ELEC_CURRENT, T_WRITE, T_READ>: sc_module{
  public:
  sca_eln::sca_terminal in,out;
  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options){}
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion ELEC_VOLTAGE->ELEC_VOLTAGE ////////////////////////////////////////////////////////////////////

// Will never be instantiated

template <typename T_WRITE, typename T_READ>
class converter<ELEC_VOLTAGE, ELEC_VOLTAGE, T_WRITE, T_READ>: sc_module{
  public:
  sca_eln::sca_terminal in,out;
  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options){}
};


/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 5 conversion from ELEC_CURRENT ***************************************************************************
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion ELEC_CURRENT->SC /////////////////////////////////////////////////////////////////////////////

template <typename T_WRITE, typename T_READ>
class converter<ELEC_CURRENT, SC, T_WRITE, T_READ>: sc_module{

  public:

  sca_eln::sca_terminal in;
  sc_out<T_READ> out;

  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options)
  {
    ostringstream warn_str;
    warn_str << "converter channel " << get_parent()->name() << ":" << endl;
    warn_str << "When the writing side is an electrical network," << endl;
    warn_str << "the writing data type MUST be set to double!!!" << endl;
    SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
  }
};

template <typename T_READ>
class converter<ELEC_CURRENT, SC, double, T_READ>: sc_module{

  sca_eln::sca_de_isink* iconv;
  datconv_sc<double, T_READ>* dconv;

  sc_signal<double> sig;
  sca_eln::sca_node_ref gnd;

  sca_eln::sca_r* terminator;  // only needed if terminating
  sca_eln::sca_node node; // resistance is set

  public:

  sca_eln::sca_terminal in;
  sc_out<T_READ> out;


  converter(sc_core::sc_module_name n, converterchannel_options<double, T_READ>* options)
  {
    if(options->elec_node_neg!=0 && options->terminating_resistance!=0.)
    {
      ostringstream warn_str;
      warn_str << "converter channel " << get_parent()->name() << ":" << endl;
      warn_str << "It is not allowed to set a terminating resistance _and_" << endl;
      warn_str << "connect the negative electrical node at the same time!" << endl;
      SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
    }

    iconv = new sca_eln::sca_de_isink("iconv");
      iconv->scale = options->elec_scaling_factor;
      iconv->p(in);
      iconv->outp(sig);
      if(options->elec_node_neg!=0) iconv->n(*(options->elec_node_neg));
      else if(options->terminating_resistance!=0.)
      {
        iconv->n(node);
        terminator = new sca_eln::sca_r("terminator");
        terminator->value = options->terminating_resistance;
        terminator->p(node);
        terminator->n(gnd);
      }
      else iconv->n(gnd);

    dconv = new datconv_sc<double, T_READ>("dconv", options);
      dconv->in(sig);
      dconv->out(out);
  }
};

template <>	// specialization for T_READ = double => no datatype conversion needed
class converter<ELEC_CURRENT, SC, double, double>: sc_module{

  sca_eln::sca_de_isink* iconv;

  sca_eln::sca_r* terminator;  // only needed if terminating
  sca_eln::sca_node node; // resistance is set

  public:

  sca_eln::sca_terminal in;
  sc_out<double> out;
  sca_eln::sca_node_ref gnd;

  converter(sc_core::sc_module_name n, converterchannel_options<double, double>* options)
  {
    if(options->elec_node_neg!=0 && options->terminating_resistance!=0.)
    {
      ostringstream warn_str;
      warn_str << "converter channel " << get_parent()->name() << ":" << endl;
      warn_str << "It is not allowed to set a terminating resistance _and_" << endl;
      warn_str << "connect the negative electrical node at the same time!" << endl;
      SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
    }
    iconv = new sca_eln::sca_de_isink("iconv");
      iconv->scale = options->elec_scaling_factor;
      iconv->p(in);
      iconv->outp(out);
      if(options->elec_node_neg!=0) iconv->n(*(options->elec_node_neg));
      else if(options->terminating_resistance!=0.)
      {
        iconv->n(node);
        terminator = new sca_eln::sca_r("terminator");
        terminator->value = options->terminating_resistance;
        terminator->p(node);
        terminator->n(gnd);
      }
      else iconv->n(gnd);
  }
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion ELEC_CURRENT->tdf /////////////////////////////////////////////////////////////////////////////

template <typename T_WRITE, typename T_READ>
class converter<ELEC_CURRENT, TDF, T_WRITE, T_READ>: sc_module{

  public:

  sca_eln::sca_terminal in;
  sca_tdf::sca_out<T_READ> out;

  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options){
    ostringstream warn_str;
    warn_str << "converter channel " << get_parent()->name() << ":" << endl;
    warn_str << "When the writing side is an electrical network," << endl;
    warn_str << "the writing data type MUST be set to double!!!" << endl;
    SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
  }
};

template <typename T_READ>
class converter<ELEC_CURRENT, TDF, double, T_READ>: sc_module{

  sca_eln::sca_de_isink* iconv;
  datconv_tdf<double, T_READ>* dconv;

  sca_tdf::sca_signal<double> sig;
  sca_eln::sca_node_ref gnd;

  sca_eln::sca_r* terminator;  // only needed if terminating
  sca_eln::sca_node node; // resistance is set

  public:

  sca_eln::sca_terminal in;
  sca_tdf::sca_out<T_READ> out;

  converter(sc_core::sc_module_name n, converterchannel_options<double, T_READ>* options)
  {
    if(options->elec_node_neg!=0 && options->terminating_resistance!=0.)
    {
      ostringstream warn_str;
      warn_str << "converter channel " << get_parent()->name() << ":" << endl;
      warn_str << "It is not allowed to set a terminating resistance _and_" << endl;
      warn_str << "connect the negative electrical node at the same time!" << endl;
      SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
    }

    iconv = new sca_eln::sca_de_isink("iconv");
      iconv->scale = options->elec_scaling_factor;
      iconv->p(in);
      iconv->outp(sig);
      if(options->elec_node_neg!=0) iconv->n(*(options->elec_node_neg));
      else if(options->terminating_resistance!=0.)
      {
        iconv->n(node);
        terminator = new sca_eln::sca_r("terminator");
        terminator->value = options->terminating_resistance;
        terminator->p(node);
        terminator->n(gnd);
      }
      else iconv->n(gnd);

    dconv = new datconv_tdf<double, T_READ>("dconv", options);
      dconv->in(sig);
      dconv->out(out);
  }
};


template <>	// specialization for T_READ = double => no datatype conversion needed
class converter<ELEC_CURRENT, TDF, double, double>: sc_module{

  sca_eln::sca_de_isink* iconv;
  sca_eln::sca_node_ref gnd;

  sca_eln::sca_r* terminator;  // only needed if terminating
  sca_eln::sca_node node; // resistance is set

  public:

  sca_eln::sca_terminal in;
  sc_out<double> out;

  converter(sc_core::sc_module_name n, converterchannel_options<double, double>* options)
  {
    if(options->elec_node_neg!=0 && options->terminating_resistance!=0.)
    {
      ostringstream warn_str;
      warn_str << "converter channel " << get_parent()->name() << ":" << endl;
      warn_str << "It is not allowed to set a terminating resistance _and_" << endl;
      warn_str << "connect the negative electrical node at the same time!" << endl;
      SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
    }
    iconv = new sca_eln::sca_de_isink("iconv");
      iconv->scale = options->elec_scaling_factor;
      iconv->p(in);
      iconv->outp(out);
      if(options->elec_node_neg!=0) iconv->n(*(options->elec_node_neg));
      else if(options->terminating_resistance!=0.)
      {
        iconv->n(node);
        terminator = new sca_eln::sca_r("terminator");
        terminator->value = options->terminating_resistance;
        terminator->p(node);
        terminator->n(gnd);
      }
      else iconv->n(gnd);
  }
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion ELEC_CURRENT->FIFO ////////////////////////////////////////////////////////////////////////////

template <typename T_WRITE, typename T_READ>
class converter<ELEC_CURRENT, FIFO, T_WRITE, T_READ>: sc_module{

  public:

  sca_eln::sca_terminal in;
  sc_fifo_out<T_READ> out;

  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options)
  {
    ostringstream warn_str;
    warn_str << "converter channel " << get_parent()->name() << ":" << endl;
    warn_str << "When the writing side is an electrical network," << endl;
    warn_str << "the writing data type MUST be set to double!!!" << endl;
    SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
  }
};

template <typename T_READ>
class converter<ELEC_CURRENT, FIFO, double, T_READ>: sc_module{

  sca_eln::sca_de_isink* iconv;
  converter<SC, FIFO, double, T_READ>* sc_fifo_conv;

  sc_signal<double> sig;
  sca_eln::sca_node_ref gnd;

  public:

  sca_eln::sca_terminal in;
  sc_fifo_out<T_READ> out;

  converter(sc_core::sc_module_name n, converterchannel_options<double, T_READ>* options)
  {
    iconv = new sca_eln::sca_de_isink("iconv");
      iconv->scale = options->elec_scaling_factor;
      iconv->p(in);
      iconv->outp(sig);
      if(options->elec_node_neg!=0) iconv->n(*(options->elec_node_neg));
      else iconv->n(gnd);

    sc_fifo_conv = new converter<SC, FIFO, double, T_READ>("sc_fifo_conv", options);

      sc_fifo_conv->in(sig);
      sc_fifo_conv->out(out);
  }
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion ELEC_CURRENT->ELEC_CURRENT ////////////////////////////////////////////////////////////////////

// Will never be instantiated

template <typename T_WRITE, typename T_READ>
class converter<ELEC_CURRENT, ELEC_CURRENT, T_WRITE, T_READ>: sc_module{
  public:
  sca_eln::sca_terminal in,out;
  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options){}
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// convertion ELEC_CURRENT->ELEC_VOLTAGE ////////////////////////////////////////////////////////////////////

// Will never be instantiated

template <typename T_WRITE, typename T_READ>
class converter<ELEC_CURRENT, ELEC_VOLTAGE, T_WRITE, T_READ>: sc_module{
  public:
  sca_eln::sca_terminal in,out;
  converter(sc_core::sc_module_name n, converterchannel_options<T_WRITE, T_READ>* options){}
};

} // end namespace TUV_ams_lib

#endif // AHES_CONVERTER_CONVERTERCHANNEL_H
// kate: space-indent on; indent-width 2; replace-tabs on;
