//-------------------------------------------------------------------------------------------
// Copyright (c) 2008-2010, Institute of Computer Technology, University of Technology Vienna
// Authors:
//   Christoph Grimm, Jan Haase, Markus Damm, Jiong Ou,  
//   Joseph Wenninger, Peter Brunmayr, Muhammad Farooq
// Contact:
//   tuv-ams-lib@ict.tuwien.ac.at
// For the license please read LICENSE.txt
//-------------------------------------------------------------------------------------------

/*
There is one possible define for template debugging: AHES_CONVERTERS_TRACE_CONV, debug will be on stderr
*/


#include <limits>
#include <cassert>

#include <boost/numeric/conversion/bounds.hpp>

#include <boost/limits.hpp>

#include<iostream>
#include<iomanip>
#include<string>
#include<cmath>
#include <iostream>
#include <sstream>
#include "utils/TUV_ams_lib_demangle.h"

#define CONVERTERS_TRACE_CONV_WORK(DESCR,T_FROM,T_TO,V_FROM,V_TO) \
CONVERTERS_TRACE_CONV_DO(cerr<<"conv ("<<DESCR<<") for "<< (demangle<T_FROM>())<<" --> "<< (demangle<T_TO>())<<" has been called ("<< V_FROM << "  --> " << V_TO <<")"<<endl;)

#define CONVERTERS_TRACE_CONV_WORK_FIXED(DESCR,T_FROM,T_TO,V_FROM,V_TO) \
CONVERTERS_TRACE_CONV_DO(cerr<<"conv ("<<DESCR<<") for "<<#T_FROM<<" --> "<<(demangle<T_TO>())<<" has been called ("<< V_FROM << "  --> " << V_TO <<")"<<endl;)

#define CONVERTERS_TRACE_CONV_WORK_TO_FIXED(DESCR,T_FROM,T_TO,V_FROM,V_TO) \
CONVERTERS_TRACE_CONV_DO(cerr<<"conv ("<<DESCR<<") for "<<(demangle<T_FROM>())<<" --> "<<#T_TO<<" has been called ("<< V_FROM << "  --> " << V_TO <<")"<<endl;)



#define CONVERTERS_TRACE_CONV_INSTANTIATED(BASE_SIG,THIS) \
CONVERTERS_TRACE_CONV_DO(cerr<<"datconv<"<< BASE_SIG << "> | "<<do_demangle(typeid(THIS).name())<<"  has been instantiated"<<endl;)
using namespace std;

namespace TUV_ams_lib {

// some defines taking the maximum values of some data types

enum Float2IntPol {F2IPOL_NONE,ROUND, FLOOR, CEIL}; // Float 2 Int Conversion Policies
enum OverflowPol {MODULO, CLIP}; // Overflow Policies

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// dummy datatype as defaults for the       /////////////////////////////////////////////////////////////////
// template-parameters in converterchannel /////////////////////////////////////////////////////////////////

namespace impl {

  template <int N>
  class dummy{	// this class will _never_ be instantiated; doing so would probably result in an error
    public:	// since the code here is minimal for a class to be accepted as datatype in SystemC
    dummy(){
	  }
    dummy(int x){
	  }
    bool operator == (const dummy &d){return true;};			 // this is the minimal code needed
    friend ostream & operator << (ostream & s, const dummy &d) {return s;} // for a new datatype to work
  };									 // within SystemC
									 // the sc_trace function is needed
  template <int N>							 // for the new datatype to be
  inline void sc_trace(sc_trace_file*, const dummy<N>&, const string& ){}; // traceble (at least formally)
} // namespace impl

enum LOG2HIGHEST_MODE {LOG2HIGHEST_FROM=0,LOG2HIGHEST_TO=1,LOG2HIGHEST_TO_FROM_VEC=2};
//enum LOG2HIGHEST_MODE {LOG2HIGHEST_FROM=false,LOG2HIGHEST_TO=true};
enum SCALINGMAX_MODE {SCALINGMAX_FROM=false,SCALINGMAX_TO=true};

enum tdf_FIFO_OPT {error, constant, hold, discardOldest, discardCurrent}; // options for the handling of full resp empty internal buffers when converting from tdf to FIFO or vice versa


SC_MODULE(tdf_clock){

  sc_out<bool> out;

  sc_time clk_period;
  sc_event start_it;

  SC_CTOR(tdf_clock){
    SC_THREAD(do_it);
    sensitive << start_it;
  }

  void start(sc_time t)
  {
    clk_period = t;
    start_it.notify();
  }

  void do_it(){
    wait();
    while(true)
    {
      out.write(true);
      wait(clk_period/2);
      out.write(false);
      wait(clk_period/2);
    }
  }

};


template <typename T_RANGE,typename T_WRITE>
struct converterchannel_options_base{

  bool use_lost_val_alerter;
  double elec_scaling_factor;
  tdf_clock* tdf_clk_gen;
  bool rangeScaling; 		// Flag if certain data type conversions use also range scaling
  T_RANGE rangeMax,rangeMin,rangeTargetMax, rangeTargetMin; 	// the bounds of values to be scaled
  tdf_FIFO_OPT FIFO2tdfemptybuffer; // option for handling empty buffers when converting fifos to tdf
  tdf_FIFO_OPT tdf2FIFOfullbuffer; // option for handling full buffers when converting tdf to fifos
  tdf_FIFO_OPT SC2FIFOfullbuffer; // option for handling full buffers when converting sc to fifos
  tdf_FIFO_OPT ELEC2FIFOfullbuffer; // option for handling full buffers when converting elec_x to fifos
  double terminating_resistance;
  sca_eln::sca_node* elec_node_neg;
  Float2IntPol f2ipol;
  OverflowPol oflowpol;

  void init()
  {
    rangeMax=scalingMax(static_cast<T_RANGE>(0));
    rangeMin=scalingMin(static_cast<T_RANGE>(0));
    rangeScaling = false;
    tdf_clk_gen = 0;
    use_lost_val_alerter = false;
    elec_scaling_factor = 1.0;
    FIFO2tdfemptybuffer=error;
    tdf2FIFOfullbuffer=error;
    SC2FIFOfullbuffer=error;
    ELEC2FIFOfullbuffer=error;
    terminating_resistance = 0.;
    elec_node_neg = 0;
    f2ipol = ROUND;
    oflowpol = MODULO;
  }

  template<typename T>
  bool signedQ(T x){return true;}

  template<int N>
  bool signedQ(sc_uint<N> x){return false;}
  template<int N>
  bool signedQ(sc_biguint<N> x){return false;}
  template<int N>
  bool signedQ(sc_bv<N> x){return false;}
  template<int N>
  bool signedQ(sc_lv<N> x){return false;}
  bool signedQ(unsigned int x){return false;}
  bool signedQ(unsigned short int x){return false;}
  bool signedQ(unsigned long int x){return false;}

  bool signedQ(bool x){return false;}
  bool signedQ(sc_logic x){return false;}
  bool signedQ(sc_bit x){return false;}

#ifdef SC_INCLUDE_FX
  template <int W,int I,sc_q_mode Q, sc_o_mode O, int N>
  bool signedQ(sc_ufixed<W,I,Q,O,N> x) {return false;}
  template <int W,int I,sc_q_mode Q, sc_o_mode O, int N>
  bool signedQ(sc_ufixed_fast<W,I,Q,O,N> x) {return false;}
#endif

  template<typename T>
  int log2highest(enum LOG2HIGHEST_MODE, T x)
  {
    double y = static_cast<double>(boost::numeric::bounds<T>::highest());
    return static_cast<int>(ceil(log(y)/log(2.)));
  }

#ifdef SC_INCLUDE_FX
  inline int fixed_log2highest(enum LOG2HIGHEST_MODE mode, int W,int I) const {
    if (I==0) return 1;
    if (I<0) return -1;
    if (mode == LOG2HIGHEST_FROM)
    {
      if (W==I) return I;
      return I+1;

    }
    else if (mode==LOG2HIGHEST_TO_FROM_VEC)
    {
      return I;
    }
    else
    {
      if (W==I) return I-1;
      return I-1;
    }
  }
  template <int W,int I,sc_q_mode Q, sc_o_mode O, int N>
  int log2highest(enum LOG2HIGHEST_MODE mode,sc_fixed<W,I,Q,O,N> x) {
    CONVERTERS_TRACE_CONV_DO(cerr<<"log2highest(sc_fixed)"<<endl;)
    return fixed_log2highest(mode, W,I);
  }
  template <int W,int I,sc_q_mode Q, sc_o_mode O, int N>
  int log2highest(enum LOG2HIGHEST_MODE mode,sc_fixed_fast<W,I,Q,O,N> x) {
    CONVERTERS_TRACE_CONV_DO(cerr<<"log2highest(sc_fixed_fast)"<<endl;)
    return fixed_log2highest(mode,W,I);
  }

  inline int ufixed_log2highest(int W,int I) const {
    if (I==0) return 1;
    if (I<0) return -1;
//warning check again
    if (W==I) return I-1;
    return I;
//??for the TO CASE ?
  }
  template <int W,int I,sc_q_mode Q, sc_o_mode O, int N>
  int log2highest(enum LOG2HIGHEST_MODE,sc_ufixed<W,I,Q,O,N> x) {
    CONVERTERS_TRACE_CONV_DO(cerr<<"log2highest(sc_ufixed)"<<endl;)
    return ufixed_log2highest(W,I);
  }
  template <int W,int I,sc_q_mode Q, sc_o_mode O, int N>
  int log2highest(enum LOG2HIGHEST_MODE,sc_ufixed_fast<W,I,Q,O,N> x) {
    CONVERTERS_TRACE_CONV_DO(cerr<<"log2highest(sc_ufixed_fast)"<<endl;)
    return ufixed_log2highest(W,I);
  }
#endif

/*  int log2highest(long double x)
  {
    long double y = boost::numeric::bounds<long double>::highest();
    int i = 0;
    while(y>1)
    {
      y/=2;
      i++;
    }
    return i;
  }*/

  template<int N>
  int log2highest(enum LOG2HIGHEST_MODE,sc_uint<N> x){return N;}
  template<int N>
  int log2highest(enum LOG2HIGHEST_MODE,sc_biguint<N> x){return N;}
  template<int N>
  int log2highest(enum LOG2HIGHEST_MODE,sc_bv<N> x){return N;}
  template<int N>
  int log2highest(enum LOG2HIGHEST_MODE,sc_lv<N> x){return N;}

  template<int N>
  int log2highest(enum LOG2HIGHEST_MODE mode ,sc_int<N> x){
    if (mode==LOG2HIGHEST_TO_FROM_VEC) return N;
    return (N-1);
  }
  template<int N>
  int log2highest(enum LOG2HIGHEST_MODE mode,sc_bigint<N> x){
    if (mode==LOG2HIGHEST_TO_FROM_VEC) return N;
    return (N-1);
  }

  int log2highest(enum LOG2HIGHEST_MODE,bool x){return 0;}
  int log2highest(enum LOG2HIGHEST_MODE,sc_logic x){return 0;}
  int log2highest(enum LOG2HIGHEST_MODE,sc_bit x){return 0;}


  template<typename T>
  bool scalingChecker(T x){return false;}

  template<int N>
  bool scalingChecker(sc_int<N> x){return true;}

  template<int N>
  bool scalingChecker(sc_uint<N> x){return true;}

  template<int N>
  bool scalingChecker(sc_bigint<N> x){return true;}

  template<int N>
  bool scalingChecker(sc_biguint<N> x){return true;}

  template<int N>
  bool scalingChecker(sc_bv<N> x){return true;}

  template<int N>
  bool scalingChecker(sc_lv<N> x){return true;}

#ifdef SC_INCLUDE_FX
  template <int W,int I,sc_q_mode Q, sc_o_mode O, int N>
  bool scalingChecker(sc_ufixed<W,I,Q,O,N> x) {
    CONVERTERS_TRACE_CONV_DO(cerr<<"scalingChecker(sc_ufixed)"<<endl;)
    return true;
  }
  template <int W,int I,sc_q_mode Q, sc_o_mode O, int N>
  bool scalingChecker(sc_ufixed_fast<W,I,Q,O,N> x) {
    CONVERTERS_TRACE_CONV_DO(cerr<<"scalingChecker(sc_ufixed_fast)"<<endl;)
    return true;
  }
  template <int W,int I,sc_q_mode Q, sc_o_mode O, int N>
  bool scalingChecker(sc_fixed<W,I,Q,O,N> x) {
    CONVERTERS_TRACE_CONV_DO(cerr<<"scalingChecker(sc_fixed)"<<endl;)
    return true;
  }
  template <int W,int I,sc_q_mode Q, sc_o_mode O, int N>
  bool scalingChecker(sc_fixed_fast<W,I,Q,O,N> x) {
    CONVERTERS_TRACE_CONV_DO(cerr<<"scalingChecker(sc_fixed_fast)"<<endl;)
    return true;
  }
#endif

/*  template<typename T>
  T unsigned_scalingMax(int n){return static_cast<T_WRITE>(pow(2.,static_cast<double>(n))-1.);}*/

  T_RANGE unsigned_scalingMax(int n){
    if (n<64) {
      return static_cast<T_RANGE>(pow((double)2.,static_cast<double>(n))-1.);
    } else {
      T_RANGE tmp=1;
        while (n>63) {
        tmp=tmp*static_cast<T_RANGE>(pow(2.,static_cast<double>(63)));
        n-=63;
      }
      if (n>0) {
        tmp=tmp*static_cast<T_RANGE>(pow(2.,static_cast<double>(n)));
      }
      tmp-=static_cast<T_RANGE>(1);
      return tmp;
    }
  }


  T_RANGE signed_scalingMax(int n){
    if (n<64) {
      return static_cast<T_RANGE>(pow((double)2.,static_cast<double>(n-1))-1.);
    } else {
       n-=1; //signed  has half  the positiv area of unsigned.
      T_RANGE tmp=1;
        while (n>63) {
        tmp=tmp*static_cast<T_RANGE>(pow(2.,static_cast<double>(63)));
        n-=63;
      }
      if (n>0) {
        tmp=tmp*static_cast<T_RANGE>(pow(2.,static_cast<double>(n)));
      }
      tmp-=static_cast<T_RANGE>(1);
      return tmp;
    }
  }


  template<typename T>
  T_RANGE scalingMax(T x){return static_cast<T_RANGE>(boost::numeric::bounds<T>::highest());}

#ifdef SC_INCLUDE_FX
  T_RANGE scalingMaxUFixed(int bits,int intBits){
    sc_ufix tmp(bits,intBits);
    tmp=0;
    tmp=~tmp;
    //std::cerr<<"scalingMaxFixed("<<bits<<","<<intBits<<"):"<<tmp<<std::endl;
    return T_RANGE(tmp);
  }


  template<typename T_RANGE1,typename SRC>
    T_RANGE1 fixup_for_int(T_RANGE1,SRC) {
      std::cerr<<"fixup_for_int returning 0"<<endl;
      return static_cast<T_RANGE1>(0);
    }

#define FIXUP_FOR_INT(X)  template<typename SRC> \
    X fixup_for_int(X,SRC) { \
      /*std::cerr<<"fixup_for_int returning 1"<<endl;*/ \
      return static_cast<X>(1); \
    };

FIXUP_FOR_INT(int)
FIXUP_FOR_INT(unsigned int)
FIXUP_FOR_INT(short)
FIXUP_FOR_INT(unsigned short)
FIXUP_FOR_INT(long int)
FIXUP_FOR_INT(unsigned long int)
#undef FIXUP_FOR_INT
  /*sc_fix*/ T_RANGE scalingMaxFixed(int bits,int intBits){
    sc_ufix tmp1(bits,intBits);
    tmp1=0;
    tmp1=~tmp1;
    sc_fix tmp(bits,intBits,SC_RND_MIN_INF,SC_SAT_SYM);
    tmp=tmp1;
    //std::cerr<<"scalingMaxFixed("<<bits<<","<<intBits<<"):"<<tmp<<std::endl;
    //std::cerr<<"scalingMaxFixed("<<bits<<","<<intBits<<"):after cast to write type:"<<T_RANGE(tmp)- fixup_for_int(T_RANGE(0),tmp)<<std::endl;
    return T_RANGE(tmp)- fixup_for_int(T_RANGE(0),tmp);
  }
  T_RANGE scalingMinFixed(int bits,int intBits){
    sc_ufix tmp1(bits,intBits);
    tmp1=0;
    tmp1=~tmp1;
    sc_fix tmp(bits,intBits,SC_RND_INF,SC_SAT_SYM);
    tmp=-tmp1;
    //std::cerr<<"scalingMaxUFixed("<<bits<<","<<intBits<<"):"<<tmp<<std::endl;
    return T_RANGE(tmp);
  }
#endif

#ifdef SC_INCLUDE_FX
  template <int W,int I,sc_q_mode Q, sc_o_mode O, int N>
  T_RANGE scalingMax(sc_ufixed<W,I,Q,O,N> x) {return scalingMaxUFixed(W,I);}
  template <int W,int I,sc_q_mode Q, sc_o_mode O, int N>
  T_RANGE scalingMax(sc_ufixed_fast<W,I,Q,O,N> x) {return scalingMaxUFixed(W,I);}
  template <int W,int I,sc_q_mode Q, sc_o_mode O, int N>
  T_RANGE scalingMax(sc_fixed<W,I,Q,O,N> x) {return scalingMaxFixed(W,I);}
  template <int W,int I,sc_q_mode Q, sc_o_mode O, int N>
  T_RANGE scalingMax(sc_fixed_fast<W,I,Q,O,N> x) {return scalingMaxFixed(W,I);}
#endif
  template<int N>
  T_RANGE scalingMax(sc_uint<N> x){return unsigned_scalingMax(N);}

  template<int N>
  T_RANGE scalingMax(sc_int<N> x){return signed_scalingMax(N);}

  template<int N>
  T_RANGE scalingMax(sc_biguint<N> x){return unsigned_scalingMax(N);}

  template<int N>
  T_RANGE scalingMax(sc_bigint<N> x){return signed_scalingMax(N);}

  template<int N>
  T_RANGE scalingMax(sc_bv<N> x){return unsigned_scalingMax(N);}

  template<int N>
  T_RANGE scalingMax(sc_lv<N> x){return unsigned_scalingMax(N);}

  T_RANGE scalingMax(sc_logic x){return static_cast<T_RANGE>(1);}

  template<typename T>
  T_RANGE scalingMin(T x){return static_cast<T_RANGE>(boost::numeric::bounds<T>::lowest());}

#ifdef SC_INCLUDE_FX
  template <int W,int I,sc_q_mode Q, sc_o_mode O, int N>
  T_RANGE scalingMin(sc_ufixed<W,I,Q,O,N> x) {return T_RANGE(static_cast<sc_ufixed<W,I,Q,O,N> >(0));}
  template <int W,int I,sc_q_mode Q, sc_o_mode O, int N>
  T_RANGE scalingMin(sc_ufixed_fast<W,I,Q,O,N> x) {return T_RANGE(static_cast<sc_ufixed_fast<W,I,Q,O,N> >(0));}
  template <int W,int I,sc_q_mode Q, sc_o_mode O, int N>
  T_RANGE scalingMin(sc_fixed<W,I,Q,O,N> x) {return T_RANGE(scalingMinFixed(W,I));}
  template <int W,int I,sc_q_mode Q, sc_o_mode O, int N>
  T_RANGE scalingMin(sc_fixed_fast<W,I,Q,O,N> x) {return T_RANGE(scalingMinFixed(W,I));}
#endif

  template<int N>
  T_RANGE scalingMin(sc_uint<N> x){return static_cast<T_RANGE>(0);}

  template<int N>
  T_RANGE scalingMin(sc_int<N> x){return -(signed_scalingMax(N)+T_RANGE(1));}

  template<int N>
  T_RANGE scalingMin(sc_biguint<N> x){return static_cast<T_RANGE>(0);}

  template<int N>
  T_RANGE scalingMin(sc_bigint<N> x){return -(signed_scalingMax(N)+T_RANGE(1));}

  template<int N>
  T_RANGE scalingMin(sc_bv<N> x){return static_cast<T_RANGE>(0);}

  template<int N>
  T_RANGE scalingMin(sc_lv<N> x){return static_cast<T_RANGE>(0);}

  T_RANGE scalingMin(sc_logic x){return static_cast<T_RANGE>(0);}


};




template <typename T_WRITE, typename T_READ>
struct converterchannel_options: public converterchannel_options_base<T_WRITE,T_WRITE>{

  T_READ const_val_empty_fifo;

  converterchannel_options()
  {
    converterchannel_options_base<T_WRITE,T_WRITE>::init();
    T_WRITE tw = static_cast<T_WRITE>(0);
    T_READ tr = static_cast<T_READ>(0);

    if(log2highest(LOG2HIGHEST_FROM,tw) <= log2highest(LOG2HIGHEST_TO,tr))  // =>  positive input stay smaller than output data type range
    {					   // then we set the upper output range to the upper input range
      //std::cerr<<"log2highest(tw);log2highest(tr)"<<log2highest(LOG2HIGHEST_FROM,tw)<<";"<<log2highest(LOG2HIGHEST_TO,tr)<<std::endl;

      converterchannel_options_base<T_WRITE,T_WRITE>::rangeTargetMax=scalingMax(tw);
      if(signedQ(tr))   		   // if the ouput data type is signed we can do the same for the
        converterchannel_options_base<T_WRITE,T_WRITE>::rangeTargetMin=scalingMin(tw);	// lower range
      else    				   // otherwise the lower range can savely set to 0 = tw
        converterchannel_options_base<T_WRITE,T_WRITE>::rangeTargetMin=tw;
    }
    else
    {
      //std::cerr<<"log2highest(tw)>log2highest(tr)"<<log2highest(LOG2HIGHEST_FROM,tw)<<";"<<log2highest(LOG2HIGHEST_TO,tr)<<std::endl;
/*      if ( (log2highest(LOG2HIGHEST_FROM,tw)==11) && (log2highest(LOG2HIGHEST_TO,tr)==10) ) {
        std:cerr<<"BREAK HERE"<<std::endl;
      }*/
      //std::cerr<<"max:"<<scalingMax(tr)<<std::endl;
      //std::cerr<<"min:"<<scalingMin(tr);
      converterchannel_options_base<T_WRITE,T_WRITE>::rangeTargetMax=scalingMax(tr);
      if(signedQ(tw))
        converterchannel_options_base<T_WRITE,T_WRITE>::rangeTargetMin=scalingMin(tr);
      else
        converterchannel_options_base<T_WRITE,T_WRITE>::rangeTargetMin=tw;
    }

    CONVERTERS_TRACE_CONV_DO(cerr << "Range data type " << demangle<T_READ>() << " : [ " << (converterchannel_options_base<T_WRITE,T_WRITE>::rangeTargetMin) << " , " << (converterchannel_options_base<T_WRITE,T_WRITE>::rangeTargetMax) << " ]" << endl;)
    const_val_empty_fifo = tr;
  }

  void setRangeScaling(T_WRITE min, T_WRITE max)
  {
    if(scalingChecker(static_cast<T_READ>(0)))
    {
      converterchannel_options_base<T_WRITE,T_WRITE>::rangeScaling=true;
      converterchannel_options_base<T_WRITE,T_WRITE>::rangeMax=max;
      converterchannel_options_base<T_WRITE,T_WRITE>::rangeMin=min;
    }
    else
    {
      ostringstream warn_str1;
      warn_str1 << "value range scaling has no effect on conversion to data type " << demangle<T_READ>() << "!" << endl;
      SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str1.str().c_str());
    }
  }
};



#define DUMMY_CONVERTERCHANNEL_OPTIONS(WRITE,READ,TMPL_P) \
template <TMPL_P> \
struct converterchannel_options<WRITE, READ > { \
  bool use_lost_val_alerter; \
  double elec_scaling_factor; \
  tdf_clock* tdf_clk_gen; \
  bool rangeScaling; \
  tdf_FIFO_OPT FIFO2tdfemptybuffer; \
  tdf_FIFO_OPT tdf2FIFOfullbuffer; \
  tdf_FIFO_OPT SC2FIFOfullbuffer; \
  tdf_FIFO_OPT ELEC2FIFOfullbuffer; \
  double terminating_resistance;\
  READ const_val_empty_fifo; \
 sca_eln::sca_node* elec_node_neg; \
  Float2IntPol f2ipol; \
  OverflowPol oflowpol; \
 \
  converterchannel_options() \
  { \
  } \
}

#define DUMMY_CONVERTERCHANNEL_OPTIONS2(WRITE,READ,TMPL_P,TMPL_P2) \
template <TMPL_P,TMPL_P2> \
struct converterchannel_options<WRITE, READ > { \
  bool use_lost_val_alerter; \
  double elec_scaling_factor; \
  tdf_clock* tdf_clk_gen; \
  bool rangeScaling; \
  tdf_FIFO_OPT FIFO2tdfemptybuffer; \
  tdf_FIFO_OPT tdf2FIFOfullbuffer; \
  tdf_FIFO_OPT SC2FIFOfullbuffer; \
  tdf_FIFO_OPT ELEC2FIFOfullbuffer; \
  double terminating_resistance;\
  READ const_val_empty_fifo; \
 sca_eln::sca_node* elec_node_neg; \
  Float2IntPol f2ipol; \
  OverflowPol oflowpol; \
 \
  converterchannel_options() \
  { \
  } \
}


#ifdef SC_INCLUDE_FX

#define DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(WRITE,READ) \
template <int W, int I, sc_q_mode Q, sc_o_mode O, int N> \
struct converterchannel_options<WRITE<W,I,Q,O,N>, READ > { \
  bool use_lost_val_alerter; \
  double elec_scaling_factor; \
  tdf_clock* tdf_clk_gen; \
  bool rangeScaling; \
  tdf_FIFO_OPT FIFO2tdfemptybuffer; \
  tdf_FIFO_OPT tdf2FIFOfullbuffer; \
  tdf_FIFO_OPT SC2FIFOfullbuffer; \
  tdf_FIFO_OPT ELEC2FIFOfullbuffer; \
  double terminating_resistance;\
  READ const_val_empty_fifo; \
 sca_eln::sca_node* elec_node_neg; \
  Float2IntPol f2ipol; \
  OverflowPol oflowpol; \
 \
  converterchannel_options() \
  { \
  } \
}

#define DUMMY_CONVERTERCHANNEL_FLOATDOUBLE_FIXED_OPTIONS(T_WRITE,T_READ) \
template <int W, int I, sc_q_mode Q, sc_o_mode O, int N> \
struct converterchannel_options<T_WRITE,T_READ<W,I,Q,O,N> >: \
  public converterchannel_options_base<T_WRITE,T_WRITE > { \
 \
  converterchannel_options() \
  { \
    converterchannel_options_base<T_WRITE,T_WRITE>::init();\
    T_WRITE tw = static_cast<T_WRITE>(0);\
    T_READ<W,I,Q,O,N>  tr = static_cast<T_READ<W,I,Q,O,N> >(0);   \
    \
    if(log2highest(LOG2HIGHEST_FROM,tw) <=  log2highest(LOG2HIGHEST_TO,tr)) \
    { \
      /*std::cerr<<"log2highest(tw);log2highest(tr)"<<log2highest(LOG2HIGHEST_FROM,tw)<<";"<<log2highest(LOG2HIGHEST_TO,tr)<<std::endl;*/ \
\
      converterchannel_options_base<T_WRITE,T_WRITE>::rangeTargetMax=scalingMax(tw); \
      if(signedQ(tr)) \
        converterchannel_options_base<T_WRITE,T_WRITE>::rangeTargetMin=scalingMin(tw); \
      else              \
        converterchannel_options_base<T_WRITE,T_WRITE>::rangeTargetMin=tw;              \
    }    \
    else \
    {   \
      /*std::cerr<<"log2highest(tw)>log2highest(tr)"<<log2highest(LOG2HIGHEST_FROM,tw)<<";"<<log2highest(LOG2HIGHEST_TO,tr)<<std::endl;*/ \
      /*std::cerr<<"max:"<<scalingMax(tr)<<std::endl;*/ \
      /*std::cerr<<"min:"<<scalingMin(tr);*/ \
      converterchannel_options_base<T_WRITE,T_WRITE>::rangeTargetMax=scalingMax(tr);      \
      if(signedQ(tw))        \
        converterchannel_options_base<T_WRITE,T_WRITE>::rangeTargetMin=scalingMin(tr);  \
      else \
        converterchannel_options_base<T_WRITE,T_WRITE>::rangeTargetMin=tw; \
    }    \
    \
    CONVERTERS_TRACE_CONV_DO(cerr << "Range data type " << (demangle<T_READ<W,I,Q,O,N> >()) << " : [ " << (converterchannel_options_base<T_WRITE,T_WRITE>::rangeTargetMin) << " , " << (converterchannel_options_base<T_WRITE,T_WRITE>::rangeTargetMax) << " ]" << endl;) \
    /*const_val_empty_fifo = tr;*/ \
  } \
 \
}

#endif

#define DUMMY_CONVERTERCHANNEL_SC_INT_OPTIONS(WRITE,READ) \
template <int N1, int W, int I, sc_q_mode Q, sc_o_mode O, int N> \
struct converterchannel_options<WRITE<N1>,READ<W,I,Q,O,N> > { \
  bool use_lost_val_alerter; \
  double elec_scaling_factor; \
  tdf_clock* tdf_clk_gen; \
  bool rangeScaling; \
  tdf_FIFO_OPT FIFO2tdfemptybuffer; \
  tdf_FIFO_OPT tdf2FIFOfullbuffer; \
  tdf_FIFO_OPT SC2FIFOfullbuffer; \
  tdf_FIFO_OPT ELEC2FIFOfullbuffer; \
  double terminating_resistance;\
  READ<W,I,Q,O,N> const_val_empty_fifo; \
 sca_eln::sca_node* elec_node_neg; \
  Float2IntPol f2ipol; \
  OverflowPol oflowpol; \
 \
  converterchannel_options() \
  { \
  } \
}




DUMMY_CONVERTERCHANNEL_OPTIONS2(sc_bv<N>,impl::dummy<N1>,int N,int N1);
DUMMY_CONVERTERCHANNEL_OPTIONS2(T_WRITE,impl::dummy<N>,typename T_WRITE,int N);
DUMMY_CONVERTERCHANNEL_OPTIONS2(sc_lv<N1>,impl::dummy<N>,int N1,int N);
//DUMMY_CONVERTERCHANNEL_OPTIONS(T_WRITE,impl::dummy<2>,typename T_WRITE);
DUMMY_CONVERTERCHANNEL_OPTIONS(sc_bv<N>,sc_bit,int N);
DUMMY_CONVERTERCHANNEL_OPTIONS(sc_bv<N>,sc_logic,int N);
DUMMY_CONVERTERCHANNEL_OPTIONS(sc_bv<N>,bool,int N);
DUMMY_CONVERTERCHANNEL_OPTIONS(sc_bit,bool,);
DUMMY_CONVERTERCHANNEL_OPTIONS(sc_bit,sc_bit,);
DUMMY_CONVERTERCHANNEL_OPTIONS(sc_bit,sc_logic,);
DUMMY_CONVERTERCHANNEL_OPTIONS(sc_bit,impl::dummy<N>,int N);
DUMMY_CONVERTERCHANNEL_OPTIONS(sc_logic,sc_bit,);
DUMMY_CONVERTERCHANNEL_OPTIONS(sc_logic,sc_logic,);
DUMMY_CONVERTERCHANNEL_OPTIONS(sc_logic,sc_lv<N>,int N);
DUMMY_CONVERTERCHANNEL_OPTIONS(sc_logic,double,);
DUMMY_CONVERTERCHANNEL_OPTIONS(T_WRITE,bool,typename T_WRITE);
DUMMY_CONVERTERCHANNEL_OPTIONS(T_WRITE,sc_logic,typename T_WRITE);
DUMMY_CONVERTERCHANNEL_OPTIONS(T_WRITE,sc_bit,typename T_WRITE);
DUMMY_CONVERTERCHANNEL_OPTIONS(sc_bit,T_READ,typename T_READ);

#ifdef SC_INCLUDE_FX
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_fixed,bool);
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_fixed_fast,bool);
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_ufixed,bool);
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_ufixed_fast,bool);

DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_fixed,float);
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_fixed_fast,float);
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_ufixed,float);
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_ufixed_fast,float);

DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_fixed,double);
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_fixed_fast,double);
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_ufixed,double);
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_ufixed_fast,double);


DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_fixed,sc_logic);
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_fixed_fast,sc_logic);
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_ufixed,sc_logic);
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_ufixed_fast,sc_logic);
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_fixed,sc_bit);
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_fixed_fast,sc_bit);
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_ufixed,sc_bit);
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_ufixed_fast,sc_bit);
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_fixed,impl::dummy<1>);
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_fixed_fast,impl::dummy<1>);
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_ufixed,impl::dummy<1>);
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_ufixed_fast,impl::dummy<1>);
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_fixed,impl::dummy<2>);
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_fixed_fast,impl::dummy<2>);
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_ufixed,impl::dummy<2>);
DUMMY_CONVERTERCHANNEL_FIXED_BOOL_OPTIONS(sc_ufixed_fast,impl::dummy<2>);

DUMMY_CONVERTERCHANNEL_FLOATDOUBLE_FIXED_OPTIONS(float,sc_fixed);
DUMMY_CONVERTERCHANNEL_FLOATDOUBLE_FIXED_OPTIONS(float,sc_ufixed);
DUMMY_CONVERTERCHANNEL_FLOATDOUBLE_FIXED_OPTIONS(float,sc_fixed_fast);
DUMMY_CONVERTERCHANNEL_FLOATDOUBLE_FIXED_OPTIONS(float,sc_ufixed_fast);

DUMMY_CONVERTERCHANNEL_FLOATDOUBLE_FIXED_OPTIONS(double,sc_fixed);
DUMMY_CONVERTERCHANNEL_FLOATDOUBLE_FIXED_OPTIONS(double,sc_ufixed);
DUMMY_CONVERTERCHANNEL_FLOATDOUBLE_FIXED_OPTIONS(double,sc_fixed_fast);
DUMMY_CONVERTERCHANNEL_FLOATDOUBLE_FIXED_OPTIONS(double,sc_ufixed_fast);

// DUMMY_CONVERTERCHANNEL_SC_INT_OPTIONS(sc_int,sc_fixed);
// DUMMY_CONVERTERCHANNEL_SC_INT_OPTIONS(sc_int,sc_ufixed);
// DUMMY_CONVERTERCHANNEL_SC_INT_OPTIONS(sc_int,sc_fixed_fast);
// DUMMY_CONVERTERCHANNEL_SC_INT_OPTIONS(sc_int,sc_ufixed_fast);
// DUMMY_CONVERTERCHANNEL_SC_INT_OPTIONS(sc_uint,sc_fixed);
// DUMMY_CONVERTERCHANNEL_SC_INT_OPTIONS(sc_uint,sc_ufixed);
// DUMMY_CONVERTERCHANNEL_SC_INT_OPTIONS(sc_uint,sc_fixed_fast);
// DUMMY_CONVERTERCHANNEL_SC_INT_OPTIONS(sc_uint,sc_ufixed_fast);
// DUMMY_CONVERTERCHANNEL_SC_INT_OPTIONS(sc_bigint,sc_fixed);
// DUMMY_CONVERTERCHANNEL_SC_INT_OPTIONS(sc_bigint,sc_ufixed);
// DUMMY_CONVERTERCHANNEL_SC_INT_OPTIONS(sc_bigint,sc_fixed_fast);
// DUMMY_CONVERTERCHANNEL_SC_INT_OPTIONS(sc_bigint,sc_ufixed_fast);
// DUMMY_CONVERTERCHANNEL_SC_INT_OPTIONS(sc_biguint,sc_fixed);
// DUMMY_CONVERTERCHANNEL_SC_INT_OPTIONS(sc_biguint,sc_ufixed);
// DUMMY_CONVERTERCHANNEL_SC_INT_OPTIONS(sc_biguint,sc_fixed_fast);
// DUMMY_CONVERTERCHANNEL_SC_INT_OPTIONS(sc_biguint,sc_ufixed_fast);
#endif

template <int N, typename T_READ>
struct converterchannel_options<sc_bv<N>, T_READ> : public converterchannel_options_base<long double,long double>{

  sc_bv<N> rangeMax,rangeMin,rangeTargetMax, rangeTargetMin; 	// the bounds of values to be scaled
  T_READ const_val_empty_fifo;
  int targlog2highest_c;

  converterchannel_options()
  {
    converterchannel_options<sc_bv<N>,T_READ>::init();
    sc_bv<N> tw = static_cast<sc_bv<N> >(false);	// sets all bits to 0
    T_READ tr = static_cast<T_READ>(0);
    rangeMax = static_cast<sc_bv<N> >(true);	// sets all bits to 1
    rangeMin = tw;
    rangeTargetMin = tw;

    int l2htr = log2highest(LOG2HIGHEST_TO,tr);
    targlog2highest_c=log2highest(LOG2HIGHEST_TO_FROM_VEC,tr);
    if(log2highest(LOG2HIGHEST_FROM,tw) <= l2htr)
      rangeTargetMax = static_cast<sc_bv<N> >(true); 	// sets all bits to 1, no overflow possible
    else
    {
      //rangeTargetMax = static_cast<sc_bv<N> >(static_cast<long int>(pow(2.,static_cast<double>(log2highest(tr)))-1.));
      rangeTargetMax = static_cast<sc_bv<N> >(false);
      for(int i = 0; i < l2htr; i++)
      {
        rangeTargetMax[i] = sc_dt::Log_1;
      }
      //cout << "log2highest von " << typeid(T_READ).name() << " : " << log2highest(tr) << " => " << pow(2.,static_cast<double>(log2highest(tr)))-1. << endl;
    }
    cout << "Range data type " << demangle<T_READ>() << " : [ " << rangeTargetMin << " , " << rangeTargetMax << " ]" << endl;
    const_val_empty_fifo = static_cast<T_READ>(0);
  }

  void setRangeScaling(sc_bv<N> min, sc_bv<N> max)
  {
    ostringstream warn_str1;
    warn_str1 << "value range scaling has no effect on conversion from data type sc_bv<" << N << "> !" << endl;
    SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str1.str().c_str());
  }
};


template <int N, typename T_READ>
struct converterchannel_options<sc_lv<N>, T_READ>{

  bool use_lost_val_alerter;
  double elec_scaling_factor;
  tdf_clock* tdf_clk_gen;
  bool rangeScaling; 		// Flag if certain data type conversions use also range scaling
  tdf_FIFO_OPT FIFO2tdfemptybuffer; // option for handling empty buffers when converting fifos to tdf
  tdf_FIFO_OPT tdf2FIFOfullbuffer; // option for handling full buffers when converting tdf to fifos
  tdf_FIFO_OPT SC2FIFOfullbuffer;
  tdf_FIFO_OPT ELEC2FIFOfullbuffer;
  double terminating_resistance;
  T_READ const_val_empty_fifo;
 sca_eln::sca_node* elec_node_neg;
  Float2IntPol f2ipol;
  OverflowPol oflowpol;

  void setRangeScaling(sc_lv<N> min, sc_lv<N> max)
  {
    ostringstream warn_str1;
    warn_str1 << "value range scaling has no effect on conversion from data type sc_lv<" << N << "> !" << endl;
    SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str1.str().c_str());
  }

  converterchannel_options()
  {
    rangeScaling = false;
    tdf_clk_gen = 0;
    use_lost_val_alerter = false;
    elec_scaling_factor = 1.0;
    FIFO2tdfemptybuffer=error;
    tdf2FIFOfullbuffer=error;
    SC2FIFOfullbuffer=error;
    ELEC2FIFOfullbuffer=error;
    terminating_resistance = 0.;
    const_val_empty_fifo = static_cast<T_READ>(0);
    elec_node_neg = 0;
    f2ipol = ROUND;
    oflowpol = MODULO;
  }
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// general conversion via static cast	  ///////////////////////////////////////////////////////////////////

// this standard converter will be specialized when needed (which will almost always be the case)


template <typename T_WRITE, typename T_READ>
class datconv{

  public:

  datconv(converterchannel_options<T_WRITE, T_READ>* options){};

  T_READ conv(T_WRITE x){
    CONVERTERS_TRACE_CONV_WORK("T_WRITE,T_READ",T_WRITE,T_READ,x,static_cast<T_READ>(x));
    return static_cast<T_READ>(x);
  };
};


/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// conversion from bool, sc_bit, sc_logic  //////////////////////////////////////////////////////////////////


// everything else not covered in this section is covered by standard converter


template <>
class datconv<bool, bool>{

  public:

  datconv(converterchannel_options<bool, bool>* options){};

  bool conv(bool x){return x;};
};

// the standart cast from bool, sc_bit and sc_logic to sc_bv<N> and sc_lv<N> causes _all_ entries in the
// vector to be set to the value of the castee. This is chenged here such only the LSB will be set to
// that value.

template <int N>
class datconv<bool, sc_bv<N> >{

  public:

  datconv(converterchannel_options<bool, sc_bv<N> >* options){};

  sc_bv<N> conv(bool x)
  {
    if(x)
      return static_cast<sc_bv<N> >(1);
    else
      return static_cast<sc_bv<N> >(0);
  };
};

template <int N>
class datconv<bool, sc_lv<N> >{

  public:

  datconv(converterchannel_options<bool, sc_lv<N> >* options){};

  sc_lv<N> conv(bool x)
  {
    if(x)
      return static_cast<sc_lv<N> >(1);
    else
      return static_cast<sc_lv<N> >(0);
  };
};


template <>
class datconv<sc_bit, sc_bit>{

  public:

  datconv(converterchannel_options<sc_bit, sc_bit>* options){};

  sc_bit conv(sc_bit x){return x;};
};


template <int N>
class datconv<sc_bit, sc_bv<N> >{

  public:

  datconv(converterchannel_options<sc_bit, sc_bv<N> >* options){};

  sc_bv<N> conv(sc_bit x)
  {
    if(static_cast<bool>(x))
      return static_cast<sc_bv<N> >(1);
    else
      return static_cast<sc_bv<N> >(0);
  };
};

template <int N>
class datconv<sc_bit, sc_lv<N> >{

  public:

  datconv(converterchannel_options<sc_bit, sc_lv<N> >* options){};

  sc_lv<N> conv(sc_bit x)
  {
    if(static_cast<bool>(x))
      return static_cast<sc_lv<N> >(1);
    else
      return static_cast<sc_lv<N> >(0);
  };
};

template <>
class datconv<sc_logic, sc_logic>{

  public:

  datconv(converterchannel_options<sc_logic, sc_logic>* options){};

  sc_logic conv(sc_logic x){return x;};
};

template <int N>
class datconv<sc_logic, sc_lv<N> >{

  public:

  datconv(converterchannel_options<sc_logic, sc_lv<N> >* options){};

  sc_lv<N> conv(sc_logic x)
  {
    sc_lv<N> res = static_cast<sc_lv<N> >(0);
    res.set_bit(0,x.value());
    return res;
  };
};

// all other conversions from sc_logic are forbidden, and this converter will produce an
// elaboration time error

template <typename T>
struct ASSERT_CONVERSION_FROM_SC_LOGIC_TO_SPECIFIED_DATATYPE_IS_NOT_ALLOWED__ILLEGAL_TYPE_IS
{
};


struct dummy_illegal_type {};
template<>
struct ASSERT_CONVERSION_FROM_SC_LOGIC_TO_SPECIFIED_DATATYPE_IS_NOT_ALLOWED__ILLEGAL_TYPE_IS<dummy_illegal_type>
{
  static void COMPILE_ERROR_BECAUSE_OF_MISSING_MEMBER_EXPECTED() {}
};



template <typename T>
class datconv<sc_logic, T>{

  public:

  datconv(converterchannel_options<sc_logic, T>* options)
  {
    ASSERT_CONVERSION_FROM_SC_LOGIC_TO_SPECIFIED_DATATYPE_IS_NOT_ALLOWED__ILLEGAL_TYPE_IS<T>::COMPILE_ERROR_BECAUSE_OF_MISSING_MEMBER_EXPECTED();
    ostringstream warn_str;
    warn_str << "ERROR: conversion from data type sc_logic to datatype " << endl;
    warn_str << "       " << demangle<T>() <<" is not allowed with converter channels!" << endl;
    SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
  };

  T conv(sc_logic x)
  {
    return static_cast<T>(0);
  };
};

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
Needed for ELEC_VOLTAGE, ELEC_CURRENT, currently only runtime error
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
template <>
class datconv<sc_logic, double>{

  public:

  datconv(converterchannel_options<sc_logic, double>* options)
  {
/*    ASSERT_CONVERSION_FROM_SC_LOGIC_TO_SPECIFIED_DATATYPE_IS_NOT_ALLOWED__ILLEGAL_TYPE_IS<T>::COMPILE_ERROR_BECAUSE_OF_MISSING_MEMBER_EXPECTED();*/
    ostringstream warn_str;
    warn_str << "ERROR: conversion from data type sc_logic to datatype " << endl;
    warn_str << "       " << demangle<double>() <<" is not allowed with converter channels!" << endl;
    SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
  };

  double conv(sc_logic x)
  {
    return static_cast<double>(0);
  };
};



/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// conversion from float, double, long double  //////////////////////////////////////////////////////////////

// this is the base class of all conversions from floating point types. It covers rounding, clipping and
// value range scaling.

template <typename T,typename T_READ>
class datconv_float_scaling_base
{
  public:

  bool rangeScaling;
  T rangeMin, rangeMax, myMin, myMax, scaleFactor;
  Float2IntPol f2ipol;
  OverflowPol oflowpol;

  datconv_float_scaling_base(Float2IntPol _f2ipol, OverflowPol _oflowpol, bool _rangeScaling, T _rangeMin, T _rangeMax, T _myMin, T _myMax)
  {
    f2ipol = _f2ipol;
    oflowpol = _oflowpol;
    rangeScaling = _rangeScaling;
    rangeMin = _rangeMin; rangeMax = _rangeMax;
    myMax=_myMax; myMin=_myMin;
    if(rangeScaling)
    {
      scaleFactor=(myMax-myMin)/(rangeMax-rangeMin);
    }
    else
    {
      scaleFactor=static_cast<T>(1);
    }

  };

  T round(T x)
  {
    if(x>=0)
      return floor(x + 0.5);
    else
      return ceil(x - 0.5);
  }

  T conv_base(T x){
    T val;
    if(rangeScaling)
    {
      if(x < rangeMin || x > rangeMax)
      {
        ostringstream warn_str;
        warn_str << "Input value " << x << " out of bounds of previously defined value range scale [ " << rangeMin << " , " << rangeMax << " ] !" << endl;
        SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str.str().c_str());
	if(oflowpol==CLIP)
	{
	  if(x>rangeMax) val = (rangeMax-rangeMin)*scaleFactor+myMin;
	  else val = (rangeMin-rangeMin)*scaleFactor+myMin;
	}
	else val = (x-rangeMin)*scaleFactor+myMin;
      }
      else val = (x-rangeMin)*scaleFactor+myMin;
    }
    else
    {
      if(x < myMin || x > myMax)
      {
        ostringstream warn_str;
        warn_str << "Input value " << x << " out of bounds of target data type " << demangle<T_READ>() << " !" << endl;
        SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str.str().c_str());
	if(oflowpol==CLIP)
	{
	  if(x>myMax) val = myMax;
	  else val = myMin;
	}
	else val = x;
      }
      else val = x;
    }
    if(f2ipol==ROUND) val = round(val);
    else if(f2ipol==FLOOR) val = floor(val);
    else if (f2ipol==CEIL) val = ceil(val);
    return val;
  }

};


template <typename T_WRITE, typename T_READ>
class datconv_float_scaling_intermed : datconv_float_scaling_base<T_WRITE,T_READ>{

  public:

  datconv_float_scaling_intermed(converterchannel_options<T_WRITE, T_READ>* options): datconv_float_scaling_base<T_WRITE,T_READ>(
    options->f2ipol,
    options->oflowpol,
    options->rangeScaling,
    options->rangeMin,
    options->rangeMax,
    options->rangeTargetMin,
    options->rangeTargetMax){}

  T_READ conv(T_WRITE x){
    return static_cast<T_READ>(conv_base(x));
  }
};


template <>
class datconv_float_scaling_intermed<double,float> : datconv_float_scaling_base<double,float>{

  public:

  datconv_float_scaling_intermed(converterchannel_options<double, float>* options): datconv_float_scaling_base<double,float>(
    F2IPOL_NONE,
    options->oflowpol,
    false,
    options->rangeMin,
    options->rangeMax,
    options->rangeTargetMin,
    options->rangeTargetMax){}

  float conv(double x){
    return static_cast<float>(conv_base(x));
  }
};


template <typename T_WRITE, typename T_READ, int N>		// intermediate class for conversion to sc_bv<N> / sc_lv<N>
class datconv_float_scaling_blv_intermed : datconv_float_scaling_base<T_WRITE,T_READ>{

  public:

  datconv_float_scaling_blv_intermed(converterchannel_options<T_WRITE, T_READ>* options): datconv_float_scaling_base<T_WRITE,T_READ>(
    options->f2ipol,
    options->oflowpol,
    options->rangeScaling,
    options->rangeMin,
    options->rangeMax,
    options->rangeTargetMin,
    options->rangeTargetMax){}

  T_READ conv(T_WRITE x){
    return static_cast<T_READ>(static_cast<sc_biguint<N> >(conv_base(x)));
  }
};

template <typename T>
class datconv<float, T > : public datconv_float_scaling_intermed<float, T>{

  public:
  datconv(converterchannel_options<float, T>* options): datconv_float_scaling_intermed<float,T>(options){}
};


template <typename T>
class datconv<double, T> : public datconv_float_scaling_intermed<double, T>{

  public:
  datconv(converterchannel_options<double, T>* options): datconv_float_scaling_intermed<double, T>(options){}
};



// template <typename T>
// class datconv<long double, T> : public datconv_float_scaling_intermed<long double, T>{
//
//   public:
//   datconv(converterchannel_options<long double, T>* options): datconv_float_scaling_intermed<long double, T>(options){}
// };


template <>
class datconv<float, bool>{
//   private:
//     float m_low;
//     float m_high;
  public:

  datconv(converterchannel_options<float, bool>* _options)/*:m_low(-1),m_high(1)*/{};
  bool conv(float x)
  {
//     if ( (x<m_low) || (x>m_high) ) return true;
    return static_cast<bool>(x);
  };
};

template <>
class datconv<double, bool>{
//   private:
//     double m_low;
//     double m_high;
  public:

  datconv(converterchannel_options<double, bool>* _options)/*:m_low(-1.0),m_high(1.0)*/{};
  bool conv(double x)
  {
    // work around a multilib problem on amd64 64-linux in 32 bit mode
//     if ( (x<m_low) || (x>m_high) )
//       return true;
    return static_cast<bool>(x);
  };
};


template <int N>
class datconv<float, sc_bv<N> > : public datconv_float_scaling_blv_intermed<float, sc_bv<N> , N>{

  public:
  datconv(converterchannel_options<float, sc_bv<N> >* options): datconv_float_scaling_blv_intermed<float, sc_bv<N> , N>(options){}
};

template <int N>
class datconv<float, sc_lv<N> > : public datconv_float_scaling_blv_intermed<float, sc_lv<N> , N>{

  public:
  datconv(converterchannel_options<float, sc_lv<N> >* options): datconv_float_scaling_blv_intermed<float, sc_lv<N> , N>(options){}
};

template <int N>
class datconv<double, sc_bv<N> > : public datconv_float_scaling_blv_intermed<double, sc_bv<N> , N>{

  public:
  datconv(converterchannel_options<double, sc_bv<N> >* options): datconv_float_scaling_blv_intermed<double, sc_bv<N> , N>(options){}
};

template <int N>
class datconv<double, sc_lv<N> > : public datconv_float_scaling_blv_intermed<double, sc_lv<N> , N>{

  public:
  datconv(converterchannel_options<double, sc_lv<N> >* options): datconv_float_scaling_blv_intermed<double, sc_lv<N> , N>(options){}
};

// template <int N>
// class datconv<long double, sc_bv<N> > : public datconv_float_scaling_blv_intermed<long double, sc_bv<N> , N>{
//
//   public:
//   datconv(converterchannel_options<long double, sc_bv<N> >* options): datconv_float_scaling_blv_intermed<long double, sc_bv<N> , N>(options){}
// };

// template <int N>
// class datconv<long double, sc_lv<N> > : public datconv_float_scaling_blv_intermed<long double, sc_lv<N> , N>{
//
//   public:
//   datconv(converterchannel_options<long double, sc_lv<N> >* options): datconv_float_scaling_blv_intermed<long double, sc_lv<N> , N>(options){}
// };

template <>
class datconv<float, float >{

  public:
  datconv(converterchannel_options<float, float>* options){}

  float conv(float x){return x;};
};

template <>
class datconv<float, double >{

  public:
  datconv(converterchannel_options<float, double>* options){}

  double conv(float x){return static_cast<double>(x);};
};

// template <>
// class datconv<float, long double >{
//
//   public:
//   datconv(converterchannel_options<float, long double>* options){}
//
//   long double conv(float x){return static_cast<long double>(x);};
// };

template <>
class datconv<double, double >{

  public:
  datconv(converterchannel_options<double, double>* options){}

  double conv(double x){return x;};
};

// template <>
// class datconv<double, long double >{
//
//   public:
//   datconv(converterchannel_options<double, long double>* options){}
//
//   long double conv(double x){return static_cast<long double>(x);};
// };

// template <>
// class datconv<long double, long double >{
//
//   public:
//   datconv(converterchannel_options<long double, long double>* options){}
//
//   long double conv(long double x){return x;};
// };

template <>
class datconv<float, sc_bit>{

  public:

  datconv(converterchannel_options<float, sc_bit>* _options){};

  sc_bit conv(float x)
  {
    return static_cast<sc_bit>(static_cast<bool>(x));
  };
};

template <>
class datconv<float, sc_logic>{

  public:

  datconv(converterchannel_options<float, sc_logic>* _options){};

  sc_logic conv(float x)
  {
    return static_cast<sc_logic>(static_cast<bool>(x));
  };
};

template <>
class datconv<double, sc_bit>{

  public:

  datconv(converterchannel_options<double, sc_bit>* _options){};

  sc_bit conv(double x)
  {
    return static_cast<sc_bit>(static_cast<bool>(x));
  };
};

template <>
class datconv<double, sc_logic>{

  public:

  datconv(converterchannel_options<double, sc_logic>* _options){};

  sc_logic conv(double x)
  {
    return static_cast<sc_logic>(static_cast<bool>(x));
  };
};

// template <>
// class datconv<long double, sc_bit>{
//
//   public:
//
//   datconv(converterchannel_options<long double, sc_bit>* _options){};
//
//   sc_bit conv(long double x)
//   {
//     return static_cast<sc_bit>(static_cast<bool>(x));
//   };
// };

// template <>
// class datconv<long double, sc_logic>{
//
//   public:
//
//   datconv(converterchannel_options<long double, sc_logic>* _options){};
//
//   sc_logic conv(long double x)
//   {
//     return static_cast<sc_logic>(static_cast<bool>(x));
//   };
// };


#define CONVERTER_SPECIALIZATION_TO_DUMMY(T_WRITE,TEMPL_P) \
template <TEMPL_P,int N1> \
class datconv<T_WRITE, impl::dummy<N1> >{ \
  \
  public: \
  \
  datconv(converterchannel_options<T_WRITE, impl::dummy<N1> >* options){};   \
  \
  impl::dummy<N1> conv(T_WRITE x) \
  { \
    ostringstream warn_str; \
    warn_str << "Conversion to impl::dummy is being performed, check your source code, returning default constructed dummy instance" << endl; \
    SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str.str().c_str()); \
    return impl::dummy<N1>(); \
  };  \
};

CONVERTER_SPECIALIZATION_TO_DUMMY(sc_bv<N>,int N);



template <typename T_WRITE,int N>
class datconv_dummy_base {

  public:

  datconv_dummy_base(converterchannel_options<T_WRITE, impl::dummy<N> >* options){};

  impl::dummy<N> conv(T_WRITE x)
  {
    ostringstream warn_str;
    warn_str << "Conversion to impl::dummy is being performed, check your source code, returning default constructed dummy instance" << endl;
    SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str.str().c_str());
    return impl::dummy<N>();
  };
};

template <typename T_WRITE,int N>
class datconv<T_WRITE, impl::dummy<N> >:public datconv_dummy_base<T_WRITE,N> {
  public:
  datconv(converterchannel_options<T_WRITE, impl::dummy<N> >* options):datconv_dummy_base<T_WRITE,N >(options){};
};


#define DATCONV_DUMMY(TYPE) \
  template <int N> \
    class datconv<TYPE, impl::dummy<N> >:public datconv_dummy_base<TYPE,N > { \
      public:\
      datconv(converterchannel_options<TYPE, impl::dummy<N> >* options):datconv_dummy_base<TYPE,N >(options){}; \
  };

 #define DATCONV_DUMMY_N1(TYPE) \
   template <int N1,int N> \
     class datconv<TYPE, impl::dummy<N> >:public datconv_dummy_base<TYPE,N > { \
       public:\
       datconv(converterchannel_options<TYPE, impl::dummy<N> >* options):datconv_dummy_base<TYPE,N >(options){}; \
   };

DATCONV_DUMMY(unsigned short int);
DATCONV_DUMMY(short int);
DATCONV_DUMMY(unsigned int);
DATCONV_DUMMY(int);
DATCONV_DUMMY(sc_logic);
DATCONV_DUMMY(long int);
DATCONV_DUMMY(unsigned long int);
DATCONV_DUMMY(double);
DATCONV_DUMMY(float);
DATCONV_DUMMY_N1(sc_bigint<N1>);
DATCONV_DUMMY_N1(sc_biguint<N1>);
DATCONV_DUMMY_N1(sc_int<N1>);
DATCONV_DUMMY_N1(sc_uint<N1>);
DATCONV_DUMMY_N1(sc_lv<N1>);

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// conversion from integer types  	/////////////////////////////////////////////////////////////////////

template <typename T,typename T_TARGET>
class datconv_integer_scaling_base
{
  public:

  bool rangeScaling;
  T rangeMin, rangeMax, myMin, myMax;
  double scaleFactor;
  Float2IntPol f2ipol;
  OverflowPol oflowpol;

  datconv_integer_scaling_base(Float2IntPol _f2ipol, OverflowPol _oflowpol, bool _rangeScaling, T _rangeMin, T _rangeMax, T _myMin, T _myMax)
  {
    f2ipol = _f2ipol;
    oflowpol = _oflowpol;
    rangeScaling = _rangeScaling;
    rangeMin = _rangeMin;
    rangeMax = _rangeMax;
    myMax = _myMax;
    myMin = _myMin;
    if(rangeScaling)
    {
      scaleFactor=static_cast<double>(myMax-myMin)/static_cast<double>(rangeMax-rangeMin);
    }
    else
    {
      scaleFactor=1.;
    }

  };

  double round(double x)
  {
    if(x>=0)
      return floor(x + 0.5);
    else
      return ceil(x - 0.5);
  }

  T conv_base(T x){
    double val = 0.;
    T ival = static_cast<T>(0);
    if(rangeScaling)
    {
      if(x < rangeMin || x > rangeMax)
      {
        ostringstream warn_str;
        warn_str << "Input value " << x << " out of bounds of previously defined value range scale [ " << rangeMin << " , " << rangeMax << " ] !" << endl;
        SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str.str().c_str());
	if(oflowpol==CLIP)
	{
	  if(x>rangeMax) val = static_cast<double>(rangeMax-rangeMin)*scaleFactor+static_cast<double>(myMin);
	  else val = static_cast<double>(rangeMin-rangeMin)*scaleFactor+static_cast<double>(myMin);
	}
	else val = static_cast<double>(x-rangeMin)*scaleFactor+static_cast<double>(myMin);
      }
      else val = static_cast<double>(x-rangeMin)*scaleFactor+static_cast<double>(myMin);

      if(f2ipol==ROUND) val = round(val);
      else if(f2ipol==FLOOR) val = floor(val);
      else val = ceil(val);
      return static_cast<T>(val);
    }
    else
    {
      if( (x < myMin) || (x > myMax) )
      {
        ostringstream warn_str;
        warn_str << "Input value " << x << " out of bounds of target data type " << demangle<T_TARGET>() << " !" <<" Range: ["<<myMin<<","<<myMax<<"]"<<endl;
        SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str.str().c_str());
	if(oflowpol==CLIP)
	{
	  if(x>myMax) ival = myMax;
	  else ival = myMin;
	}
	else ival = x;
      }
      else ival = x;
      return ival;
    }
  }

};


template <typename T_WRITE, typename T_READ>
class datconv_integer_scaling_intermed : datconv_integer_scaling_base<T_WRITE,T_READ>{

  public:

  datconv_integer_scaling_intermed(converterchannel_options<T_WRITE, T_READ>* options): datconv_integer_scaling_base<T_WRITE,T_READ>(
    options->f2ipol,
    options->oflowpol,
    options->rangeScaling,
    options->rangeMin,
    options->rangeMax,
    options->rangeTargetMin,
    options->rangeTargetMax) {
      if (!options->rangeScaling) {
      T_WRITE tw = static_cast<T_WRITE>(0);
      T_READ tr = static_cast<T_READ>(0);
        if (options->log2highest(LOG2HIGHEST_FROM,tw) > options->log2highest(LOG2HIGHEST_TO,tr))
        {
          ostringstream warn_str1;
          warn_str1 << "Converting from " << demangle<T_WRITE>() << " to "<< demangle<T_READ>() << " may cause loss !" <<
          options->log2highest(LOG2HIGHEST_FROM,tw)<<"->"<<options->log2highest(LOG2HIGHEST_TO,tr)<<endl;
          //warn_str1 << "of information since " << N << " < " << M << " !" << endl;
          SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str1.str().c_str());
        }
        if (options->signedQ(tw) && (!options->signedQ(tr))) {
          ostringstream warn_str1;
          warn_str1 << "Converting from " << demangle<T_WRITE>() << " to "<< demangle<T_READ>() << " may cause loss !: ";
          warn_str1 << "signed -> unsigned";
          //warn_str1 << "of information since " << N << " < " << M << " !" << endl;
          SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str1.str().c_str());
        }
      }
    }

  T_READ conv(T_WRITE x){
    CONVERTERS_TRACE_CONV_WORK("dataconv_integer_scaling_intermed",T_WRITE,T_READ,x,static_cast<T_READ>(conv_base(x)))
    return static_cast<T_READ>(conv_base(x));
  }
};

template <typename T>
class datconv<char, T > : public datconv_integer_scaling_intermed<char, T>{

  public:
  datconv(converterchannel_options<char, T>* options): datconv_integer_scaling_intermed<char,T>(options)
  {
    CONVERTERS_TRACE_CONV_INSTANTIATED("char,T",this)
  }
};

template <typename T>
class datconv<unsigned short int, T > : public datconv_integer_scaling_intermed<unsigned short int, T>{

  public:
  datconv(converterchannel_options<unsigned short int, T>* options): datconv_integer_scaling_intermed<unsigned short int,T>(options){}
};

template <typename T>
class datconv<short int, T > : public datconv_integer_scaling_intermed<short int, T>{

  public:
  datconv(converterchannel_options<short int, T>* options): datconv_integer_scaling_intermed<short int,T>(options){}
};

template <typename T>
class datconv<unsigned int, T > : public datconv_integer_scaling_intermed<unsigned int, T>{

  public:
  datconv(converterchannel_options<unsigned int, T>* options): datconv_integer_scaling_intermed<unsigned int,T>(options){}
};

template <typename T>
class datconv<int, T > : public datconv_integer_scaling_intermed<int, T>{

  public:
  datconv(converterchannel_options<int, T>* options): datconv_integer_scaling_intermed<int,T>(options){}
};

template <typename T>
class datconv<unsigned long int, T > : public datconv_integer_scaling_intermed<unsigned long int, T>{

  public:
  datconv(converterchannel_options<unsigned long int, T>* options): datconv_integer_scaling_intermed<unsigned long int,T>(options){}
};

template <typename T>
class datconv<long int, T > : public datconv_integer_scaling_intermed<long int, T>{

  public:
  datconv(converterchannel_options<long int, T>* options): datconv_integer_scaling_intermed<long int,T>(options){}
};

template <>
class datconv<char, char> {

  public:
  datconv(converterchannel_options<char, char>* options){}

  char conv(char x){return x;};
};

template <>
class datconv<char, unsigned short int> {

  public:
  datconv(converterchannel_options<char, unsigned short int>* options){}

  unsigned int conv(char x){return static_cast<unsigned short int>(x);};
};

template <>
class datconv<char, short int> {

  public:
  datconv(converterchannel_options<char, short int>* options){}

  short int conv(char x){return static_cast<short int>(x);};
};

template <>
class datconv<char, unsigned int> {

  public:
  datconv(converterchannel_options<char, unsigned int>* options){}

  unsigned int conv(char x){return static_cast<unsigned int>(x);};
};

template <>
class datconv<char, int> {

  public:
  datconv(converterchannel_options<char, int>* options){}

  int conv(char x){return static_cast<int>(x);};
};

template <>
class datconv<char, unsigned long int> {

  public:
  datconv(converterchannel_options<char, unsigned long int>* options){}

  unsigned long int conv(char x){return static_cast<unsigned long int>(x);};
};

template <>
class datconv<char, long int> {

  public:
  datconv(converterchannel_options<char, long int>* options){}

  long int conv(char x){return static_cast<long int>(x);};
};

template <>
class datconv<char, float> {

  public:
  datconv(converterchannel_options<char, float>* options){}

  float conv(char x){return static_cast<float>(x);};
};

template <>
class datconv<char, double> {

  public:
  datconv(converterchannel_options<char, double>* options){}

  double conv(char x){return static_cast<double>(x);};
};

// template <>
// class datconv<char, long double> {
//
//   public:
//   datconv(converterchannel_options<char, long double>* options){}
//
//   long double conv(char x){return static_cast<long double>(x);};
// };

template <>
class datconv<unsigned short int, unsigned short int> {

  public:
  datconv(converterchannel_options<unsigned short int, unsigned short int>* options){}

  unsigned short int conv(unsigned short int x){return x;};
};

template <>
class datconv<unsigned short int, unsigned int> {

  public:
  datconv(converterchannel_options<unsigned short int, unsigned int>* options){}

  unsigned int conv(unsigned short int x){return static_cast<unsigned int>(x);};
};

template <>
class datconv<unsigned short int, int> {

  public:
  datconv(converterchannel_options<unsigned short int, int>* options){}

  int conv(unsigned short int x){return static_cast<int>(x);};
};

template <>
class datconv<unsigned short int, unsigned long int> {

  public:
  datconv(converterchannel_options<unsigned short int, unsigned long int>* options){}

  unsigned long int conv(unsigned short int x){return static_cast<unsigned long int>(x);};
};

template <>
class datconv<unsigned short int, long int> {

  public:
  datconv(converterchannel_options<unsigned short int, long int>* options){}

  long int conv(unsigned short int x){return static_cast<long int>(x);};
};

template <>
class datconv<unsigned short int, float> {

  public:
  datconv(converterchannel_options<unsigned short int, float>* options){}

  float conv(unsigned short int x){return static_cast<float>(x);};
};

template <>
class datconv<unsigned short int, double> {

  public:
  datconv(converterchannel_options<unsigned short int, double>* options){}

  double conv(unsigned short int x){return static_cast<double>(x);};
};

// template <>
// class datconv<unsigned short int, long double> {
//
//   public:
//   datconv(converterchannel_options<unsigned short int, long double>* options){}
//
//   long double conv(unsigned short int x){return static_cast<long double>(x);};
// };



template <>
class datconv<short int, short int> {

  public:
  datconv(converterchannel_options<short int, short int>* options){}

  short int conv(short int x){return x;};
};

template <>
class datconv<short int, int> {

  public:
  datconv(converterchannel_options<short int, int>* options){}

  int conv(short int x){return static_cast<int>(x);};
};

template <>
class datconv<short int, unsigned long int> {

  public:
  datconv(converterchannel_options<short int, unsigned long int>* options){}

  unsigned long int conv(short int x){return static_cast<unsigned long int>(x);};
};

template <>
class datconv<short int, long int> {

  public:
  datconv(converterchannel_options<short int, long int>* options){}

  long int conv(short int x){return static_cast<long int>(x);};
};

template <>
class datconv<short int, float> {

  public:
  datconv(converterchannel_options<short int, float>* options){}

  float conv(short int x){return static_cast<float>(x);};
};

template <>
class datconv<short int, double> {

  public:
  datconv(converterchannel_options<short int, double>* options){}

  double conv(short int x){return static_cast<double>(x);};
};

// template <>
// class datconv<short int, long double> {
//
//   public:
//   datconv(converterchannel_options<short int, long double>* options){}
//
//   long double conv(short int x){return static_cast<long double>(x);};
// };





template <>
class datconv<unsigned int, unsigned int> {

  public:
  datconv(converterchannel_options<unsigned int, unsigned int>* options){}

  unsigned int conv(unsigned int x){return x;};
};

template <>
class datconv<unsigned int, unsigned long int> {

  public:
  datconv(converterchannel_options<unsigned int, unsigned long int>* options){}

  unsigned long int conv(unsigned int x){return static_cast<unsigned long int>(x);};
};

template <>
class datconv<unsigned int, long int> {

  public:
  datconv(converterchannel_options<unsigned int, long int>* options){}

  long int conv(unsigned int x){return static_cast<long int>(x);};
};

template <>
class datconv<unsigned int, float> {

  public:
  datconv(converterchannel_options<unsigned int, float>* options){}

  float conv(unsigned int x){return static_cast<float>(x);};
};

template <>
class datconv<unsigned int, double> {

  public:
  datconv(converterchannel_options<unsigned int, double>* options){}

  double conv(unsigned int x){return static_cast<double>(x);};
};
/*
template <>
class datconv<unsigned int, long double> {

  public:
  datconv(converterchannel_options<unsigned int, long double>* options){}

  long double conv(unsigned int x){return static_cast<long double>(x);};
};*/




template <>
class datconv<int, int> {

  public:
  datconv(converterchannel_options<int, int>* options){}

  int conv(int x){return x;};
};

template <>
class datconv<int, long int> {

  public:
  datconv(converterchannel_options<int, long int>* options){}

  long int conv(int x){return static_cast<long int>(x);};
};

template <>
class datconv<int, float> {

  public:
  datconv(converterchannel_options<int, float>* options){}

  float conv(int x){return static_cast<float>(x);};
};

template <>
class datconv<int, double> {

  public:
  datconv(converterchannel_options<int, double>* options){}

  double conv(int x){return static_cast<double>(x);};
};

// template <>
// class datconv<int, long double> {
//
//   public:
//   datconv(converterchannel_options<int, long double>* options){}
//
//   long double conv(int x){return static_cast<long double>(x);};
// };


template <>
class datconv<unsigned long int, unsigned long int> {

  public:
  datconv(converterchannel_options<unsigned long int, unsigned long int>* options){}

  unsigned long int conv(unsigned long int x){return x;};
};


template <>
class datconv<unsigned long int, float> {

  public:
  datconv(converterchannel_options<unsigned long int, float>* options){}

  float conv(unsigned long int x){return static_cast<float>(x);};
};

template <>
class datconv<unsigned long int, double> {

  public:
  datconv(converterchannel_options<unsigned long int, double>* options){}

  double conv(unsigned long int x){return static_cast<double>(x);};
};

// template <>
// class datconv<unsigned long int, long double> {
//
//   public:
//   datconv(converterchannel_options<unsigned long int, long double>* options){}
//
//   long double conv(unsigned long int x){return static_cast<long double>(x);};
// };


template <>
class datconv<long int, long int> {

  public:
  datconv(converterchannel_options<long int, long int>* options){}

  long int conv(long int x){return x;};
};

template <>
class datconv<long int, float> {

  public:
  datconv(converterchannel_options<long int, float>* options){}

  float conv(long int x){return static_cast<float>(x);};
};

template <>
class datconv<long int, double> {

  public:
  datconv(converterchannel_options<long int, double>* options){}

  double conv(long int x){return static_cast<double>(x);};
};

// template <>
// class datconv<long int, long double> {
//
//   public:
//   datconv(converterchannel_options<long int, long double>* options){}
//
//   long double conv(long int x){return static_cast<long double>(x);};
// };

template <>
class datconv<unsigned short int, sc_logic>{

  public:

  datconv(converterchannel_options<unsigned short int, sc_logic>* _options){};

  sc_logic conv(unsigned short int x)
  {
    return static_cast<sc_logic>(static_cast<bool>(x));
  };
};

template <>
class datconv<unsigned short int, sc_bit>{

  public:

  datconv(converterchannel_options<unsigned short int, sc_bit>* _options){};

  sc_bit conv(unsigned short int x)
  {
    return static_cast<sc_bit>(static_cast<bool>(x));
  };
};


template <>
class datconv<short int, sc_logic>{

  public:

  datconv(converterchannel_options<short int, sc_logic>* _options){};

  sc_logic conv(short int x)
  {
    return static_cast<sc_logic>(static_cast<bool>(x));
  };
};

template <>
class datconv<short int, sc_bit>{

  public:

  datconv(converterchannel_options<short int, sc_bit>* _options){};

  sc_bit conv(short int x)
  {
    return static_cast<sc_bit>(static_cast<bool>(x));
  };
};

template <>
class datconv<short int, bool>{
  public:

  datconv(converterchannel_options<short int, bool>* _options){};
  bool conv(short int x)
  {
    return static_cast<bool>(x);
  };
};


template <>
class datconv<unsigned int, sc_logic>{

  public:

  datconv(converterchannel_options<unsigned int, sc_logic>* _options){};

  sc_logic conv(unsigned int x)
  {
    return static_cast<sc_logic>(static_cast<bool>(x));
  };
};

template <>
class datconv<unsigned int, sc_bit>{

  public:

  datconv(converterchannel_options<unsigned int, sc_bit>* _options){};

  sc_bit conv(unsigned int x)
  {
    return static_cast<sc_bit>(static_cast<bool>(x));
  };
};

template <>
class datconv<unsigned int, bool>{
  public:

  datconv(converterchannel_options<unsigned int, bool>* _options){};
  bool conv(unsigned int x)
  {
    return static_cast<bool>(x);
  };
};

template <>
class datconv<unsigned short int, bool>{
  public:

  datconv(converterchannel_options<unsigned short int, bool>* _options){};
  bool conv(unsigned short int x)
  {
    return static_cast<bool>(x);
  };
};


template <>
class datconv<int, sc_logic>{

  public:

  datconv(converterchannel_options<int, sc_logic>* _options){};

  sc_logic conv(int x)
  {
    return static_cast<sc_logic>(static_cast<bool>(x));
  };
};
template <>
class datconv<int, sc_bit>{

  public:

  datconv(converterchannel_options<int, sc_bit>* _options){};

  sc_bit conv(int x)
  {
    return static_cast<sc_bit>(static_cast<bool>(x));
  };
};

template <>
class datconv<int, bool>{
  public:

  datconv(converterchannel_options<int, bool>* _options){};
  bool conv(int x)
  {
    return static_cast<bool>(x);
  };
};

template <>
class datconv<unsigned long int, sc_logic>{

  public:

  datconv(converterchannel_options<unsigned long int, sc_logic>* _options){};

  sc_logic conv(unsigned long int x)
  {
    return static_cast<sc_logic>(static_cast<bool>(x));
  };
};

template <>
class datconv<unsigned long int, sc_bit>{

  public:

  datconv(converterchannel_options<unsigned long int, sc_bit>* _options){};

  sc_bit conv(unsigned long int x)
  {
    return static_cast<sc_bit>(static_cast<bool>(x));
  };
};

template <>
class datconv<unsigned long int, bool>{
  public:

  datconv(converterchannel_options<unsigned long int, bool>* _options){};
  bool conv(unsigned long int x)
  {
    return static_cast<bool>(x);
  };
};

template <>
class datconv<long int, sc_logic>{

  public:

  datconv(converterchannel_options<long int, sc_logic>* _options){};

  sc_logic conv(long int x)
  {
    return static_cast<sc_logic>(static_cast<bool>(x));
  };
};

template <>
class datconv<long int, sc_bit>{

  public:

  datconv(converterchannel_options<long int, sc_bit>* _options){};

  sc_bit conv(long int x)
  {
    return static_cast<sc_bit>(static_cast<bool>(x));
  };
};


template <>
class datconv<long, bool>{
  public:

  datconv(converterchannel_options<long, bool>* _options){};
  bool conv(long x)
  {
    return static_cast<bool>(x);
  };
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// conversion from sc_lv<N>   	/////////////////////////////////////////////////////////////////////////////

// ... are forbidden for all target data types except sc_lv<N> itself!

template <int N>
class datconv<sc_lv<N>, sc_lv<N> > {

  public:
  datconv(converterchannel_options<sc_lv<N>, sc_lv<N> >* options){}

  sc_lv<N> conv(sc_lv<N> x){return x;};
};

template <int N, int M>
class datconv<sc_lv<N>, sc_lv<M> > {

  public:
  datconv(converterchannel_options<sc_lv<N>, sc_lv<M> >* options)
  {
    if(N>M)
    {
      ostringstream warn_str1;
      warn_str1 << "Converting from sc_lv<" << N << "> to sc_lv<" << M << "> may cause loss" << endl;
      warn_str1 << "of information since " << N << " < " << M << " !" << endl;
      SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str1.str().c_str());
    }
  }

  sc_lv<M> conv(sc_lv<N> x)
  {
    if(N>M)
    {
      bool overflow = false;
      for(int i = N-1; i >= M & !overflow; i--) if(x[i]!=sc_dt::Log_0) overflow = true;
      if(overflow)
      {
        ostringstream warn_str;
        warn_str << "Input value " << x << " out of bounds of target data type sc_lv<" << M << "> !" << endl;
        SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str.str().c_str());
      }
    }
    return x;
  };
};


template <typename T>
struct ASSERT_CONVERSION_FROM_SC_LV_TO_ANYTHING_BUT_SC_LV_NOT_ALLOWED__ILLEGAL_TYPE_IS
{
};

template <int N>
struct ASSERT_CONVERSION_FROM_SC_LV_TO_ANYTHING_BUT_SC_LV_NOT_ALLOWED__ILLEGAL_TYPE_IS<sc_lv<N> >
{
  static void COMPILE_ERROR_BECAUSE_OF_MISSING_MEMBER_EXPECTED() {}
};

template <int N, typename T>
class datconv<sc_lv<N>, T>{

  public:

  datconv(converterchannel_options<sc_lv<N>, T>* _options)
  {
    ASSERT_CONVERSION_FROM_SC_LV_TO_ANYTHING_BUT_SC_LV_NOT_ALLOWED__ILLEGAL_TYPE_IS<T>::COMPILE_ERROR_BECAUSE_OF_MISSING_MEMBER_EXPECTED();
    ostringstream warn_str;
    warn_str << "ERROR: conversion from data type sc_lv<N> to any other datatype " << endl;
    warn_str << "       than sc_lv<M> is not allowed with converter channels!" << endl;
    SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
  };


  T conv(sc_lv<N> x)
  {
    return static_cast<T>(0);
  };
};

/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
Needed for ELEC_VOLTAGE, ELEC_CURRENT, currently only runtime error
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
template <int N>
class datconv<sc_lv<N>, double>{

  public:

  datconv(converterchannel_options<sc_lv<N>, double>* _options)
  {
    //ASSERT_CONVERSION_FROM_SC_LV_TO_ANYTHING_BUT_SC_LV_NOT_ALLOWED__ILLEGAL_TYPE_IS<T>::COMPILE_ERROR_BECAUSE_OF_MISSING_MEMBER_EXPECTED();
    ostringstream warn_str;
    warn_str << "ERROR: conversion from data type sc_lv<N> to any other datatype " << endl;
    warn_str << "       than sc_lv<M> is not allowed with converter channels!" << endl;
    SC_REPORT_ERROR("ANDRES-Converterchannels",warn_str.str().c_str());
  };


  double conv(sc_lv<N> x)
  {
    return static_cast<double>(0);
  };
};


/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// conversion from sc_bv<N>   	/////////////////////////////////////////////////////////////////////////////

template <int N>
class datconv<sc_bv<N>, sc_bv<N> > {

  public:
  datconv(converterchannel_options<sc_bv<N>, sc_bv<N> >* options){}

  sc_bv<N> conv(sc_bv<N> x){return x;};
};

template <int N>
class datconv<sc_bv<N>, sc_lv<N> > {

  public:
  datconv(converterchannel_options<sc_bv<N>, sc_lv<N> >* options){}

  sc_lv<N> conv(sc_bv<N> x){return static_cast<sc_lv<N> >(x);};
};

template <int N, typename T,int M>
class datconv_bv_vec_base {

  OverflowPol oflowpol;

  public:
  datconv_bv_vec_base(converterchannel_options<sc_bv<N>, T >* options)
  {
    oflowpol = options->oflowpol;
    if(N>M)
    {
      ostringstream warn_str1;
      warn_str1 << "Converting from sc_bv<" << N << "> to sc_bv<" << M << "> may cause" << endl;
      warn_str1 << "loss of information!" << endl;
      SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str1.str().c_str());
    }
  }

  T conv(sc_bv<N> x)
  {
    T ival = static_cast<T>(x);
    if(N>M)
    {
      bool overflow = false;
      for(int i = N-1; i >= M & !overflow; i--) if(x[i]==sc_dt::Log_1) overflow = true;
      if(overflow)
      {
        ostringstream warn_str;
        warn_str << "Input value " << x << " out of bounds of target data type sc_bv<" << M << "> !" << endl;
        SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str.str().c_str());
        if(oflowpol==CLIP) ival = static_cast<T>(true);
      }
    }
    return ival;
  };
};


template <int N, int M>
class datconv<sc_bv<N>, sc_bv<M> >: public datconv_bv_vec_base<N,sc_bv<M>,M > {
  public:
    datconv(converterchannel_options<sc_bv<N>, sc_bv<M> >* options):
      datconv_bv_vec_base<N,sc_bv<M>,M >(options) {}
};

template <int N, int M>
class datconv<sc_bv<N>, sc_lv<M> >: public datconv_bv_vec_base<N,sc_lv<M>,M > {
  public:
    datconv(converterchannel_options<sc_bv<N>, sc_lv<M> >* options):
      datconv_bv_vec_base<N,sc_lv<M>,M >(options) {}
};


template <int N, typename T>
class datconv<sc_bv<N>, T>{

  sc_bv<N> myMax, myMin; 	// value bounds for overflow control
  OverflowPol oflowpol;
  int targlog2highest;
  int targlog2highest_c;
  public:

  datconv(converterchannel_options<sc_bv<N>, T>* options)
  {
    myMin = options->rangeTargetMin;
    myMax = options->rangeTargetMax;
    oflowpol = options->oflowpol;
    targlog2highest_c=options->targlog2highest_c;
    targlog2highest = options->log2highest(LOG2HIGHEST_TO,static_cast<T>(0));
    int mylog2highest = options->log2highest(LOG2HIGHEST_FROM,static_cast<sc_bv<N> >(0));
    // cerr<<"BV -->>>>>>>>>>>>"<<mylog2highest<<"----"<<targlog2highest<<endl;
    if (targlog2highest<mylog2highest) {
        ostringstream warn_str1;
        warn_str1 << "Converting from sc_bv<"<<N<< ">" << " to "<< demangle<T>() << " may cause loss !" <<
        mylog2highest<<"->"<<targlog2highest<<endl;
        //warn_str1 << "of information since " << N << " < " << M << " !" << endl;
        SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str1.str().c_str());
    }
    else if(targlog2highest>mylog2highest) {
      targlog2highest = mylog2highest;
      targlog2highest_c=mylog2highest;
    }
  };

  bool greater(sc_bv<N> x, sc_bv<N> y)
  {
    bool res = false;
    bool desc = false;
    for(int i = N-1; i>=0 && !desc; i--)
    {
      if(x[i]==sc_dt::Log_1 && y[i]==sc_dt::Log_0)	// => x > y
      {
       res = true;
       desc = true;
      }
      else if(x[i]==sc_dt::Log_0 && y[i]==sc_dt::Log_1)	// => y > x
       desc = true;
    }
    return res; // if loop runs until end, x = y
  }


  T conv(sc_bv<N> x)
  {
    T val = static_cast<T>(0);
    sc_bv<N> ival = x;
    if(greater(x, myMax))
    {
      ostringstream warn_str;
      warn_str << "Input value " << x << " out of bounds of target data type " << demangle<T>() << " !" << endl;
      SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str.str().c_str());
      if(oflowpol==CLIP) ival = myMax;
    }
    for(int i = 0; i<targlog2highest_c; i++)
    {
      if(ival[i] == sc_dt::Log_1) val = val + static_cast<T>(pow(2.,static_cast<double>(i)));
      //cout<<"Val after bit("<<i<<"):"<<val<<endl;
    }
    return val;
  };
};


template <int N, typename T>
class datconv_sc_bv_to_sc_int_base{

  sc_bv<N> myMax, myMin; 	// value bounds for overflow control
  OverflowPol oflowpol;
  int targlog2highest;
  bool targetSigned;

  public:

  void init (converterchannel_options<sc_bv<N>, T>* options)
  {
    myMin = options->rangeTargetMin;
    myMax = options->rangeTargetMax;
    oflowpol = options->oflowpol;
    targlog2highest = options->log2highest(LOG2HIGHEST_TO,static_cast<T>(0));
    targetSigned = options->signedQ(static_cast<T>(0));
    if(targetSigned) targlog2highest++; //leads to 2-complement interpretation if the input
    int mylog2highest = options->log2highest(LOG2HIGHEST_FROM,static_cast<sc_bv<N> >(0));
    if(targlog2highest>mylog2highest) targlog2highest = mylog2highest;
  };

  bool greater(sc_bv<N> x, sc_bv<N> y)
  {
    bool res = false;
    bool desc = false;
    for(int i = N-1; i>=0 && !desc; i--)
    {
      if(x[i]==sc_dt::Log_1 && y[i]==sc_dt::Log_0)	// => x > y
      {
       res = true;
       desc = true;
      }
      else if(x[i]==sc_dt::Log_0 && y[i]==sc_dt::Log_1)	// => y > x
       desc = true;
    }
    return res; // if loop runs until end, x = y
  }


  T conv_base(sc_bv<N> x)
  {
    T val = static_cast<T>(0);
    sc_bv<N> ival = x;
    if(greater(x, myMax))
    {
      ostringstream warn_str;
      warn_str << "Input value " << x << " out of bounds of target data type " << demangle<T>() << " !" << endl;
      SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str.str().c_str());
      if(oflowpol==CLIP) ival = myMax;
    }
    for(int i = 0; i<targlog2highest; i++)
      val.set(i, ival.get_bit(i));
    return val;
  };
};

template <int N, int M>
class datconv<sc_bv<N>, sc_int<M> >: datconv_sc_bv_to_sc_int_base<N, sc_int<M> > {

  public:
  datconv(converterchannel_options<sc_bv<N>, sc_int<M> >* options){init(options);}
  sc_int<M> conv(sc_bv<N> x){return conv_base(x);};
};

template <int N, int M>
class datconv<sc_bv<N>, sc_uint<M> >: datconv_sc_bv_to_sc_int_base<N, sc_uint<M> > {

  public:
  datconv(converterchannel_options<sc_bv<N>, sc_uint<M> >* options){init(options);}
  sc_uint<M> conv(sc_bv<N> x){return conv_base(x);};
};

template <int N, int M>
class datconv<sc_bv<N>, sc_bigint<M> >: datconv_sc_bv_to_sc_int_base<N, sc_bigint<M> > {

  public:
  datconv(converterchannel_options<sc_bv<N>, sc_bigint<M> >* options){init(options);}
  sc_bigint<M> conv(sc_bv<N> x){return conv_base(x);};
};

template <int N, int M>
class datconv<sc_bv<N>, sc_biguint<M> >: datconv_sc_bv_to_sc_int_base<N, sc_biguint<M> > {

  public:
  datconv(converterchannel_options<sc_bv<N>, sc_biguint<M> >* options){init(options);}
  sc_biguint<M> conv(sc_bv<N> x){return conv_base(x);};
};



template <int N>
class datconv_sc_bv_to_bool_base{

  protected:

  datconv_sc_bv_to_bool_base(){}

  bool conv_base(sc_bv<N> x)
  {
    bool res = static_cast<bool>(x.to_uint()); // works for N small enough (N<=32)
    if(N<33) return res;
    for(int i=N-1; i>31 && !res; i--) res = static_cast<bool>(x.get_bit(i)); 	// otherwise we can't be sure this is realy the 0-vector
    return res;								// and have to search the uppermost bits
  }
};

template <int N>
class datconv<sc_bv<N>, bool>: datconv_sc_bv_to_bool_base<N> {

  public:
  datconv(converterchannel_options<sc_bv<N>, bool>* options){}
  bool conv(sc_bv<N> x){return conv_base(x);};
};

template <int N>
class datconv<sc_bv<N>, sc_bit>: datconv_sc_bv_to_bool_base<N> {

  public:
  datconv(converterchannel_options<sc_bv<N>, sc_bit>* options){}
  sc_bit conv(sc_bv<N> x){return static_cast<sc_bit>(conv_base(x));};
};

template <int N>
class datconv<sc_bv<N>, sc_logic>: datconv_sc_bv_to_bool_base<N> {

  public:
  datconv(converterchannel_options<sc_bv<N>, sc_logic>* options){}
  sc_logic conv(sc_bv<N> x){return static_cast<sc_logic>(conv_base(x));};
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// conversion from sc_int<N> etc.  	/////////////////////////////////////////////////////////////////////

template <int N>
class datconv<sc_int<N>, sc_int<N> > {

  public:
  datconv(converterchannel_options<sc_int<N>, sc_int<N> >* options){}

  sc_int<N> conv(sc_int<N> x){return x;};
};

template <int N>
class datconv<sc_uint<N>, sc_uint<N> > {

  public:
  datconv(converterchannel_options<sc_uint<N>, sc_uint<N> >* options){}

  sc_uint<N> conv(sc_uint<N> x){return x;};
};

template <int N>
class datconv<sc_bigint<N>, sc_bigint<N> > {

  public:
  datconv(converterchannel_options<sc_bigint<N>, sc_bigint<N> >* options){}

  sc_bigint<N> conv(sc_bigint<N> x){return x;};
};

template <int N>
class datconv<sc_biguint<N>, sc_biguint<N> > {

  public:
  datconv(converterchannel_options<sc_biguint<N>, sc_biguint<N> >* options){}

  sc_biguint<N> conv(sc_biguint<N> x){return x;};
};

template <typename T_WRITE, typename T_READ>
class datconv_sc_int_base{

  T_WRITE myMax, myMin; 	// value bounds for overflow control
  OverflowPol oflowpol;

  public:

  void init(converterchannel_options<T_WRITE, T_READ>* options)
  {
    myMin = options->rangeTargetMin;
    myMax = options->rangeTargetMax;
    oflowpol = options->oflowpol;
  };


  T_READ conv_base(T_WRITE x)
  {
    T_WRITE ival = x;
    if(x > myMax || x < myMin)
    {
      ostringstream warn_str;
      warn_str << "Input value " << x << " out of bounds of target data type " << demangle<T_READ>() << " !" << endl;
      SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str.str().c_str());
      if(oflowpol==CLIP)
      {
	if(x>myMax) ival = myMax;
	else ival = myMin;
      }
    }
    return static_cast<T_READ>(ival);
  };
};


template <typename T_WRITE, typename T_READ>
class datconv_sc_bigint_float_base{

  T_WRITE myMax, myMin; 	// value bounds for overflow control
  OverflowPol oflowpol;
  int targlog2highest;

  public:

  void init(converterchannel_options<T_WRITE, T_READ>* options)
  {
    myMin = options->rangeTargetMin;
    myMax = options->rangeTargetMax;

    //cout << "*** MyMin = " << myMin << " , myMax = " << myMax << endl;

    oflowpol = options->oflowpol;
    targlog2highest = options->log2highest(LOG2HIGHEST_TO,static_cast<T_READ>(0));
    int mylog2highest = options->log2highest(LOG2HIGHEST_FROM,static_cast<T_WRITE>(0));

    if (options->signedQ(static_cast<T_WRITE>(0)) && (!options->signedQ(static_cast<T_READ>(0))))
    {
          ostringstream warn_str1;
          warn_str1 << "Converting from " << demangle<T_WRITE>() << " to "<< demangle<T_READ>() << " may cause loss !:";
          warn_str1 << "signed -> unsigned";
          //warn_str1 << "of information since " << N << " < " << M << " !" << endl;
          SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str1.str().c_str());
    }
    else
    {
      if (mylog2highest>targlog2highest)
      {
        ostringstream warn_str1;
        warn_str1 << "Converting from " << demangle<T_WRITE>() << " to "<< demangle<T_READ>() << " may cause loss !" <<
        mylog2highest<<"->"<<targlog2highest<<endl;
        //warn_str1 << "of information since " << N << " < " << M << " !" << endl;
        SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str1.str().c_str());
      }
    }

    if (options->signedQ(static_cast<T_READ>(0)))
      targlog2highest=targlog2highest+1;

    if(targlog2highest>mylog2highest) targlog2highest = mylog2highest;
  };


  T_READ conv_base(T_WRITE x)
  {
    T_WRITE ival = x;
    T_READ val = static_cast<T_READ>(0);
    if(x > myMax || x < myMin)
    {
      ostringstream warn_str;
      warn_str << "Input value " << x << " out of bounds of target data type " << demangle<T_READ>() << " !" << endl;
      SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str.str().c_str());
      if(oflowpol==CLIP)
      {
	if(x>myMax) ival = myMax;
	else ival = myMin;
      }
    }
    bool negative = false;
    if(ival < static_cast<T_WRITE>(0))
    {
      negative = true;
      ival = - ival;
    }
    for(int i = 0; i<targlog2highest; i++)
    {
      if(ival[i]) val = val + static_cast<T_READ>(pow(2.,static_cast<double>(i)));
    }
    if(negative) val = -val;
    return val;
  };
};

template <typename T, int N>
class datconv<sc_int<N>, T>: datconv_sc_int_base<sc_int<N>, T>{

  public:
  datconv(converterchannel_options<sc_int<N>, T>* options){datconv_sc_int_base<sc_int<N>, T>::init(options);}

  T conv(sc_int<N> x){return conv_base(x);};
};

template <typename T, int N>
class datconv<sc_uint<N>, T>: datconv_sc_int_base<sc_uint<N>, T>{

  public:
  datconv(converterchannel_options<sc_uint<N>, T>* options){datconv_sc_int_base<sc_uint<N>, T>::init(options);}

  T conv(sc_uint<N> x){return conv_base(x);};
};

template <typename T, int N>
class datconv<sc_bigint<N>, T>: datconv_sc_bigint_float_base<sc_bigint<N>, T>{

  public:
  datconv(converterchannel_options<sc_bigint<N>, T>* options){datconv_sc_bigint_float_base<sc_bigint<N>, T>::init(options);}

  T conv(sc_bigint<N> x){return conv_base(x);};
};

template <typename T, int N>
class datconv<sc_biguint<N>, T>: datconv_sc_bigint_float_base<sc_biguint<N>, T>{

  public:
  datconv(converterchannel_options<sc_biguint<N>, T>* options){datconv_sc_bigint_float_base<sc_biguint<N>, T>::init(options);}

  T conv(sc_biguint<N> x){return conv_base(x);};
};

template <int N, int M>
class datconv<sc_bigint<N>, sc_bv<M> >: datconv_sc_int_base<sc_bigint<N>, sc_bv<M> >{

  public:
  datconv(converterchannel_options<sc_bigint<N>, sc_bv<M> >* options){datconv_sc_int_base<sc_bigint<N>, sc_bv<M> >::init(options);}

  sc_bv<M> conv(sc_bigint<N> x){return conv_base(x);};
};

template <int N, int M>
class datconv<sc_biguint<N>, sc_bv<M> >: datconv_sc_int_base<sc_biguint<N>, sc_bv<M> >{

  public:
  datconv(converterchannel_options<sc_biguint<N>, sc_bv<M> >* options){datconv_sc_int_base<sc_biguint<N>, sc_bv<M> >::init(options);}

  sc_bv<M> conv(sc_biguint<N> x){return conv_base(x);};
};


template <int N, int M>
class datconv<sc_bigint<N>, sc_lv<M> >: datconv_sc_int_base<sc_bigint<N>, sc_lv<M> >{

  public:
  datconv(converterchannel_options<sc_bigint<N>, sc_lv<M> >* options){datconv_sc_int_base<sc_bigint<N>, sc_lv<M> >::init(options);}

  sc_lv<M> conv(sc_bigint<N> x){return conv_base(x);};
};

template <int N, int M>
class datconv<sc_biguint<N>, sc_lv<M> >: datconv_sc_int_base<sc_biguint<N>, sc_lv<M> >{

  public:
  datconv(converterchannel_options<sc_biguint<N>, sc_lv<M> >* options){datconv_sc_int_base<sc_biguint<N>, sc_lv<M> >::init(options);}

  sc_lv<M> conv(sc_biguint<N> x){return conv_base(x);};
};



template <typename T_WRITE, int N>
class datconv_sc_int_to_bool_base{

  protected:

  datconv_sc_int_to_bool_base(){}

  bool conv_base(T_WRITE x)
  {
    bool res = static_cast<bool>(x.to_int()); // works for N small enough (N<=32)
    if(N<33) return res;
    for(int i=N-1; i>31 && !res; i--) res = static_cast<bool>(x[i]); 	// otherwise we can't be sure this is realy the 0-vector
    return res;								// and have to search the uppermost bits
  }
};

template <int N>
class datconv<sc_int<N>, bool>: datconv_sc_int_to_bool_base<sc_int<N> ,N> {

  public:
  datconv(converterchannel_options<sc_int<N>, bool>* options){}
  bool conv(sc_int<N> x){return conv_base(x);};
};

template <int N>
class datconv<sc_int<N>, sc_bit>: datconv_sc_int_to_bool_base<sc_int<N> ,N> {

  public:
  datconv(converterchannel_options<sc_int<N>, sc_bit>* options){}
  sc_bit conv(sc_int<N> x){return static_cast<sc_bit>(conv_base(x));};
};

template <int N>
class datconv<sc_int<N>, sc_logic>: datconv_sc_int_to_bool_base<sc_int<N> ,N> {

  public:
  datconv(converterchannel_options<sc_int<N>, sc_logic>* options){}
  sc_logic conv(sc_int<N> x){return static_cast<sc_logic>(conv_base(x));};
};

template <int N>
class datconv<sc_uint<N>, bool>: datconv_sc_int_to_bool_base<sc_uint<N> ,N> {

  public:
  datconv(converterchannel_options<sc_uint<N>, bool>* options){}
  bool conv(sc_uint<N> x){return conv_base(x);};
};

template <int N>
class datconv<sc_uint<N>, sc_bit>: datconv_sc_int_to_bool_base<sc_uint<N> ,N> {

  public:
  datconv(converterchannel_options<sc_uint<N>, sc_bit>* options){}
  sc_bit conv(sc_uint<N> x){return static_cast<sc_bit>(conv_base(x));};
};

template <int N>
class datconv<sc_uint<N>, sc_logic>: datconv_sc_int_to_bool_base<sc_uint<N> ,N> {

  public:
  datconv(converterchannel_options<sc_uint<N>, sc_logic>* options){}
  sc_logic conv(sc_uint<N> x){return static_cast<sc_logic>(conv_base(x));};
};






template <int N>
class datconv<sc_bigint<N>, bool>: datconv_sc_int_to_bool_base<sc_bigint<N> ,N> {

  public:
  datconv(converterchannel_options<sc_bigint<N>, bool>* options){}
  bool conv(sc_bigint<N> x){return conv_base(x);};
};

template <int N>
class datconv<sc_bigint<N>, sc_bit>: datconv_sc_int_to_bool_base<sc_bigint<N> ,N> {

  public:
  datconv(converterchannel_options<sc_bigint<N>, sc_bit>* options){}
  sc_bit conv(sc_bigint<N> x){return static_cast<sc_bit>(conv_base(x));};
};

template <int N>
class datconv<sc_bigint<N>, sc_logic>: datconv_sc_int_to_bool_base<sc_bigint<N> ,N> {

  public:
  datconv(converterchannel_options<sc_bigint<N>, sc_logic>* options){}
  sc_logic conv(sc_bigint<N> x){return static_cast<sc_logic>(conv_base(x));};
};

template <int N>
class datconv<sc_biguint<N>, bool>: datconv_sc_int_to_bool_base<sc_biguint<N> ,N> {

  public:
  datconv(converterchannel_options<sc_biguint<N>, bool>* options){}
  bool conv(sc_biguint<N> x){return conv_base(x);};
};

template <int N>
class datconv<sc_biguint<N>, sc_bit>: datconv_sc_int_to_bool_base<sc_biguint<N> ,N> {

  public:
  datconv(converterchannel_options<sc_biguint<N>, sc_bit>* options){}
  sc_bit conv(sc_biguint<N> x){return static_cast<sc_bit>(conv_base(x));};
};

template <int N>
class datconv<sc_biguint<N>, sc_logic>: datconv_sc_int_to_bool_base<sc_biguint<N> ,N> {

  public:
  datconv(converterchannel_options<sc_biguint<N>, sc_logic>* options){}
  sc_logic conv(sc_biguint<N> x){return static_cast<sc_logic>(conv_base(x));};
};



#ifdef SC_INCLUDE_FX


//BEGIN: Explicit conversion for trace/debug purposes fo identical fixed types
  //do nothing, just return the value itself, conversion between equal datatypes
  #define SC_X_FIXED_DO_NOTHING(from_to) template <int W,int I,sc_q_mode Q, sc_o_mode O,int N> \
    class datconv<from_to<W,I,Q,O,N>, from_to<W,I,Q,O,N> >{ \
      public: \
        datconv(converterchannel_options<from_to<W,I,Q,O,N>, from_to<W,I,Q,O,N> >* options)\
        {\
          CONVERTERS_TRACE_CONV_INSTANTIATED("from_to<W,I,Q,O,N>, from_to<W,I,Q,O,N>",this) \
        }; \
        from_to<W,I,Q,O,N> conv(from_to<W,I,Q,O,N> x)\
        {\
          CONVERTERS_TRACE_CONV_DO(cerr<<"conv for "<<#from_to<<" --> "<<#from_to<<" has been called ("<<x<<")"<<endl;)\
          return x;\
        }; \
    };

  SC_X_FIXED_DO_NOTHING(sc_fixed)
  SC_X_FIXED_DO_NOTHING(sc_ufixed)
  SC_X_FIXED_DO_NOTHING(sc_fixed_fast)
  SC_X_FIXED_DO_NOTHING(sc_ufixed_fast)
//END: Explicit conversion for trace/debug purposes fo identical fixed types


//BEGIN: Explicit conversion between normal and fast types
  //static casts between equivalent fast and not normal types
  #define SC_X_FIXED_Y(from,to) template <int W,int I,sc_q_mode Q, sc_o_mode O,int N> \
    class datconv<from<W,I,Q,O,N>, to<W,I,Q,O,N> >{ \
      public: \
        datconv(converterchannel_options<from<W,I,Q,O,N>, to<W,I,Q,O,N> >* options)\
        {\
          CONVERTERS_TRACE_CONV_INSTANTIATED("from<W,I,Q,O,N>, to<W,I,Q,O,N>",this) \
        }; \
        to<W,I,Q,O,N> conv(from<W,I,Q,O,N> x)\
        {\
          CONVERTERS_TRACE_CONV_DO(cerr<<"conv for "<<#from<<" --> "<<#to<<" has been called: ("<<x<<" --> " << (static_cast<to<W,I,Q,O,N> >(x)) <<")"<<endl;)\
          return static_cast<to<W,I,Q,O,N> >(x);\
        }; \
    };


  SC_X_FIXED_Y(sc_fixed,sc_fixed_fast)
  SC_X_FIXED_Y(sc_fixed_fast,sc_fixed)
  SC_X_FIXED_Y(sc_ufixed,sc_ufixed_fast)
  SC_X_FIXED_Y(sc_ufixed_fast,sc_ufixed)
//END: Explicit conversion between normal and fast types

//BEGIN: Static cast between elements of different world and interger word length
  //different word_length,integer_word_length
  #define SC_X_FIXED_Y_A_B_WL(from,to) \
    template <int Wf,int If,sc_q_mode Qf, sc_o_mode Of,int Nf, \
              int Wt,int It,sc_q_mode Qt, sc_o_mode Ot, int Nt> \
    class datconv<from<Wf,If,Qf,Of,Nf>, to<Wt,It,Qt,Ot,Nt> >{ \
      public:\
        datconv(converterchannel_options<from<Wf,If,Qf,Of,Nf>, to<Wt,It,Qt,Ot,Nt> >* options){}; \
        to<Wt,It,Qt,Ot,Nt> conv(from<Wf,If,Qf,Of,Nf> x) \
      { \
        CONVERTERS_TRACE_CONV_DO(cerr<<"conv for "<<#from<< \
          "<"<<Wf<<","<<If<<","<<Qf<<","<<Of<<","<<Nf<<">" \
          <<" --> "<<#to << \
          "<"<<Wt<<","<<It<<","<<Qt<<","<<Ot<<","<<Nt<<">" \
          <<" has been called: ("<<x<<" --> " << (static_cast<to<Wt,It,Qt,Ot,Nt> >(x)) <<")"<<endl;)\
        return static_cast<to<Wt,It,Qt,Ot,Nt> >(x); \
      }; \
    };\

  //explicit instead of implicit conversion listing
  SC_X_FIXED_Y_A_B_WL(sc_fixed,sc_fixed_fast)
  SC_X_FIXED_Y_A_B_WL(sc_fixed_fast,sc_fixed)
  SC_X_FIXED_Y_A_B_WL(sc_ufixed,sc_ufixed_fast)
  SC_X_FIXED_Y_A_B_WL(sc_ufixed_fast,sc_ufixed)

  SC_X_FIXED_Y_A_B_WL(sc_fixed,sc_fixed)
  SC_X_FIXED_Y_A_B_WL(sc_fixed_fast,sc_fixed_fast)
  SC_X_FIXED_Y_A_B_WL(sc_ufixed,sc_ufixed)
  SC_X_FIXED_Y_A_B_WL(sc_ufixed_fast,sc_ufixed_fast)

  SC_X_FIXED_Y_A_B_WL(sc_fixed,sc_ufixed)
  SC_X_FIXED_Y_A_B_WL(sc_ufixed,sc_fixed)
  SC_X_FIXED_Y_A_B_WL(sc_fixed_fast,sc_ufixed_fast)
  SC_X_FIXED_Y_A_B_WL(sc_ufixed_fast,sc_fixed_fast)

  SC_X_FIXED_Y_A_B_WL(sc_fixed,sc_ufixed_fast)
  SC_X_FIXED_Y_A_B_WL(sc_ufixed_fast,sc_fixed)
  SC_X_FIXED_Y_A_B_WL(sc_ufixed,sc_fixed_fast)
  SC_X_FIXED_Y_A_B_WL(sc_fixed_fast,sc_ufixed)
//END: Static cast between elements of different world and interger word length

//BEGIN: SC FIXED VALUES TO SC BIT/LOGIC
  //Fixed values to sc_bit/sc_logic via a bool intermediate
  //do nothing, just return the value itself, conversion between equal datatypes
  #define SC_X_FIXED_Y_BOOL(from,to) template <int W,int I,sc_q_mode Q, sc_o_mode O,int N> \
    class datconv<from<W,I,Q,O,N>, to >{ \
      public: \
        datconv(converterchannel_options<from<W,I,Q,O,N>, to >* options)\
        {\
          CONVERTERS_TRACE_CONV_INSTANTIATED(#from"<W,I,Q,O,N>, "#to,this) \
        }; \
        to conv(from<W,I,Q,O,N> x)\
        {\
          CONVERTERS_TRACE_CONV_DO(cerr<<"conv for "<<#from<<" --> "<<#to<<" has been called ("<<x<<" --> "<<static_cast<to>(static_cast<bool>(x))<<")"<<endl;)\
          return static_cast<to>(static_cast<bool>(x));\
        }; \
    };

  SC_X_FIXED_Y_BOOL(sc_fixed,bool)
  SC_X_FIXED_Y_BOOL(sc_ufixed,bool)
  SC_X_FIXED_Y_BOOL(sc_fixed_fast,bool)
  SC_X_FIXED_Y_BOOL(sc_ufixed_fast,bool)
  SC_X_FIXED_Y_BOOL(sc_fixed,sc_bit)
  SC_X_FIXED_Y_BOOL(sc_ufixed,sc_bit)
  SC_X_FIXED_Y_BOOL(sc_fixed_fast,sc_bit)
  SC_X_FIXED_Y_BOOL(sc_ufixed_fast,sc_bit)
  SC_X_FIXED_Y_BOOL(sc_fixed,sc_logic)
  SC_X_FIXED_Y_BOOL(sc_ufixed,sc_logic)
  SC_X_FIXED_Y_BOOL(sc_fixed_fast,sc_logic)
  SC_X_FIXED_Y_BOOL(sc_ufixed_fast,sc_logic)
//END: SC FIXED VALUES TO SC BIT/LOGIC

//warning TESTING
template <typename T_RANGE, typename T_WRITE, typename T_READ>
struct converterchannel_options_fixed_base: public converterchannel_options_base<T_RANGE,T_WRITE>{

  T_READ const_val_empty_fifo;

  converterchannel_options_fixed_base()
  {
    converterchannel_options_base<T_RANGE,T_WRITE>::init();
    T_WRITE  tw = static_cast<T_WRITE>(0);
    T_READ tr = static_cast<T_READ>(0);

    //std::cerr<<"######################################################## sc_(u)fixed_fast"<<std::endl;

    if(log2highest(LOG2HIGHEST_FROM,tw) <= log2highest(LOG2HIGHEST_TO,tr))  // =>  positive input stay smaller than output data type range
    {                                      // then we set the upper output range to the upper input range
      //std::cerr<<"log2highest(tw);log2highest(tr)"<<log2highest(LOG2HIGHEST_FROM,tw)<<";"<<log2highest(LOG2HIGHEST_TO,tr)<<std::endl;

      converterchannel_options_base<T_RANGE,T_WRITE >::rangeTargetMax=scalingMax(tw);
      if(signedQ(tr))                      // if the ouput data type is signed we can do the same for the
        converterchannel_options_base<T_RANGE, T_WRITE >::rangeTargetMin=scalingMin(tw);  // lower range
      else                                 // otherwise the lower range can savely set to 0 = tw
        converterchannel_options_base<T_RANGE,T_WRITE >::rangeTargetMin=tw;
    }
    else
    {
      //std::cerr<<"log2highest(tw)>log2highest(tr)"<<log2highest(LOG2HIGHEST_FROM,tw)<<";"<<log2highest(LOG2HIGHEST_TO,tr)<<std::endl;
      /*if ( (log2highest(tw)==11) && (log2highest(tr)==10) ) {
        std:cerr<<"BREAK HERE"<<std::endl;
      }*/
      converterchannel_options_base<T_RANGE, T_WRITE>::rangeTargetMax=scalingMax(tr);
      if(signedQ(tw))
        converterchannel_options_base<T_RANGE,T_WRITE>::rangeTargetMin=scalingMin(tr);
      else
        converterchannel_options_base<T_RANGE,T_WRITE>::rangeTargetMin=tw;
    }

    CONVERTERS_TRACE_CONV_DO(cerr << "Range data type " << demangle<T_READ>() << " : [ " << (converterchannel_options_base<T_RANGE,T_WRITE>::rangeTargetMin) << " , " << (converterchannel_options_base<T_RANGE,T_WRITE>::rangeTargetMax) << " ]" << endl;)
    const_val_empty_fifo = tr;
  }

  void setRangeScaling(T_WRITE min, T_WRITE max)
  {
    if(scalingChecker(static_cast<T_READ>(0)))
    {
      converterchannel_options_base<T_RANGE,T_WRITE>::rangeScaling=true;
      converterchannel_options_base<T_RANGE,T_WRITE>::rangeMax=max;
      converterchannel_options_base<T_RANGE,T_WRITE>::rangeMin=min;
    }
    else
    {
      ostringstream warn_str1;
      warn_str1 << "value range scaling has no effect on conversion to data type " << demangle<T_READ>() << "!" << endl;
      SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str1.str().c_str());
    }
  }
};

template <int W,int I,sc_q_mode Q, sc_o_mode O,int N, typename T_READ>
struct converterchannel_options<sc_fixed_fast<W,I,Q,O,N>,T_READ>: public converterchannel_options_fixed_base<sc_fixed_fast<W+1,I+1,SC_RND>,sc_fixed_fast<W,I,Q,O,N>, T_READ>{
public:
  converterchannel_options():converterchannel_options_fixed_base<sc_fixed_fast<W+1,I+1,SC_RND>,sc_fixed_fast<W,I,Q,O,N>, T_READ>() {}
};

template <int W,int I,sc_q_mode Q, sc_o_mode O,int N, typename T_READ>
struct converterchannel_options<sc_ufixed_fast<W,I,Q,O,N>,T_READ>: public converterchannel_options_fixed_base<sc_ufixed_fast<W+1,I+1,SC_RND>,sc_ufixed_fast<W,I,Q,O,N>, T_READ>{
public:
  converterchannel_options():converterchannel_options_fixed_base<sc_ufixed_fast<W+1,I+1,SC_RND>,sc_ufixed_fast<W,I,Q,O,N>, T_READ>() {}
};

template <int W,int I,sc_q_mode Q, sc_o_mode O,int N, typename T_READ>
struct converterchannel_options<sc_fixed<W,I,Q,O,N>,T_READ>: public converterchannel_options_fixed_base<sc_fixed<W+1,I+1,SC_RND>,sc_fixed<W,I,Q,O,N>, T_READ>{
public:
  converterchannel_options():converterchannel_options_fixed_base<sc_fixed<W+1,I+1,SC_RND>,sc_fixed<W,I,Q,O,N>, T_READ>() {}
};

template <int W,int I,sc_q_mode Q, sc_o_mode O,int N, typename T_READ>
struct converterchannel_options<sc_ufixed<W,I,Q,O,N>,T_READ>: public converterchannel_options_fixed_base<sc_ufixed<W+1,I+1,SC_RND>,sc_ufixed<W,I,Q,O,N>, T_READ>{
public:
  converterchannel_options():converterchannel_options_fixed_base<sc_ufixed<W+1,I+1,SC_RND>,sc_ufixed<W,I,Q,O,N>, T_READ>() {}
};



//BEGIN: SC FIXED CONVERSION TO arbitrary data types, base
//currently only a copy of datconv_float_scaling_base, will probably diverge later on
  /////////////////////////////////////////////////////////////////////////////////////////////////////////////
  // conversion from sc_fixed,sc_ufixed,sc_fixed_fast,sc_ufixed_fast //////////////////////////////////////////

  // this is the base class of all conversions from fixed point types. It covers rounding, clipping and
  // value range scaling.

  template <typename T,typename T_READ>
  class datconv_fixed_scaling_base
  {
    public:

    bool rangeScaling;
    T rangeMin, rangeMax, myMin, myMax, scaleFactor;
    Float2IntPol f2ipol;
    OverflowPol oflowpol;

    datconv_fixed_scaling_base(Float2IntPol _f2ipol, OverflowPol _oflowpol, bool _rangeScaling, T _rangeMin, T _rangeMax, T _myMin, T _myMax)
    {
      f2ipol = _f2ipol;
      oflowpol = _oflowpol;
      rangeScaling = _rangeScaling;
      rangeMin = _rangeMin; rangeMax = _rangeMax;
      myMax=_myMax; myMin=_myMin;
      if(rangeScaling)
      {
        scaleFactor=(myMax-myMin)/(rangeMax-rangeMin);
      }
      else
      {
        scaleFactor=static_cast<T>(1);
      }

    };

    T round(T x)
    {
      if(x>=0)
        return floor(x + 0.5);
      else
        return ceil(x - 0.5);
    }

    T conv_base(T x){
      //std::cerr<<"datconv_fixed_scaling_base::conv_base: "<<x<<std::endl;
      T val;
      if(rangeScaling)
      {
        if(x < rangeMin || x > rangeMax)
        {
          ostringstream warn_str;
          warn_str << "Input value " << x << " out of bounds of previously defined value range scale [ " << rangeMin << " , " << rangeMax << " ] !" << endl;
          SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str.str().c_str());
          if(oflowpol==CLIP)
          {
            if(x>rangeMax) val = (rangeMax-rangeMin)*scaleFactor+myMin;
            else val = (rangeMin-rangeMin)*scaleFactor+myMin;
          }
          else val = (x-rangeMin)*scaleFactor+myMin;
        }
        else val = (x-rangeMin)*scaleFactor+myMin;
      }
      else
      {
        if(x < myMin || x > myMax)
        {
          ostringstream warn_str;
          warn_str << "Input value " << x << " out of bounds of target data type " << demangle<T_READ>() << " !" << endl;
          SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str.str().c_str());
          if(oflowpol==CLIP)
          {
            if(x>myMax) val = myMax;
            else val = myMin;
          }
          else val = x;
        }
        else val = x;
      }
      //std::cerr<<"datconv_fixed_scaling_base::conv_base: val before rounding/FLOOR/CEIL"<<val<<std::endl;
//warning needs more testing and evaluation
      if(f2ipol==ROUND) val = round(val);
      else if(f2ipol==FLOOR) val = floor(val);
      else val = ceil(val);
      //std::cerr<<"datconv_fixed_scaling_base::conv_base: val, before return"<<val<<std::endl;
      return val;
    }

  };
//END: SC FIXED CONVERSION TO arbitrary data types, base

//BEGIN: SC FIXED CONVERSION TO arbitrary data types
  template <typename T_WRITE, typename T_READ>
  class datconv_fixed_scaling_intermed : datconv_fixed_scaling_base<T_WRITE,T_READ>{

    public:

    datconv_fixed_scaling_intermed(converterchannel_options<T_WRITE, T_READ>* options): datconv_fixed_scaling_base<T_WRITE,T_READ>(
      options->f2ipol,
      options->oflowpol,
      options->rangeScaling,
      options->rangeMin,
      options->rangeMax,
      options->rangeTargetMin,
      options->rangeTargetMax){}

    T_READ conv(T_WRITE x){
      T_READ tmp=static_cast<T_READ>(conv_base(x));
        CONVERTERS_TRACE_CONV_DO(cerr<<"conv for "<<demangle<T_WRITE>()<<" --> "<<demangle<T_READ>()<<" has been called ("<<x<<" --> "<<tmp<<")"<<endl;);
      return tmp;
    }
  };

  #define SC_X_FIXED_Y_ANYTHING(from) template <int W,int I,sc_q_mode Q,sc_o_mode O, int N, typename T> \
  class datconv<from<W,I,Q,O,N>, T > : public datconv_fixed_scaling_intermed<from<W,I,Q,O,N>, T>{ \
    public:  \
    datconv(converterchannel_options<from<W,I,Q,O,N>, T>* options): \
      datconv_fixed_scaling_intermed<from<W,I,Q,O,N>,T>(options){} \
  };

  SC_X_FIXED_Y_ANYTHING(sc_fixed)
  SC_X_FIXED_Y_ANYTHING(sc_ufixed)
  SC_X_FIXED_Y_ANYTHING(sc_fixed_fast)
  SC_X_FIXED_Y_ANYTHING(sc_ufixed_fast)
//BEGIN: SC FIXED CONVERSION TO arbitrary data types

//BEGIN: SC FIXED CONVERSION TO float, double
  #define SC_X_FIXED_Y_FLOATDOUBLE(from,to) template <int W,int I,sc_q_mode Q,sc_o_mode O, int N> \
  class datconv<from<W,I,Q,O,N>, to >{ \
  \
    public: \
  \
    datconv(converterchannel_options<from<W,I,Q,O,N>, to>* options){}; \
  \
    to conv(from<W,I,Q,O,N> x){\
      CONVERTERS_TRACE_CONV_WORK_FIXED("T_WRITE,T_READ",#from,to,x,static_cast<to>(x)); \
      return static_cast<to>(x); \
    }; \
  };

  #define FLOATDOUBLE_SC_X_FIXED_Y(from,to) template <int W,int I,sc_q_mode Q,sc_o_mode O, int N> \
  class datconv<from,to<W,I,Q,O,N> >{ \
  \
    private: \
      from m_min,m_max; \
    public: \
  \
    datconv(converterchannel_options<from,to<W,I,Q,O,N> >* options){ \
      m_min=options->rangeTargetMin; m_max=options->rangeTargetMax; \
    }; \
  \
    to<W,I,Q,O,N> conv(from x){\
      if ( (x<m_min) || (x>m_max) ) { \
          ostringstream warn_str; \
          warn_str << "Input value " << x << " out of bounds of target data type " << demangle<to<W,I,Q,O,N> >() << " !" << endl; \
          SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str.str().c_str()); \
      } \
      CONVERTERS_TRACE_CONV_WORK_TO_FIXED("T_WRITE,T_READ",from,#to,x,(static_cast<to<W,I,Q,O,N> >(x))); \
      return static_cast<to<W,I,Q,O,N> >(x); \
    }; \
  };


  SC_X_FIXED_Y_FLOATDOUBLE(sc_fixed,float)
  SC_X_FIXED_Y_FLOATDOUBLE(sc_ufixed,float)
  SC_X_FIXED_Y_FLOATDOUBLE(sc_fixed_fast,float)
  SC_X_FIXED_Y_FLOATDOUBLE(sc_ufixed_fast,float)
  SC_X_FIXED_Y_FLOATDOUBLE(sc_fixed,double)
  SC_X_FIXED_Y_FLOATDOUBLE(sc_ufixed,double)
  SC_X_FIXED_Y_FLOATDOUBLE(sc_fixed_fast,double)
  SC_X_FIXED_Y_FLOATDOUBLE(sc_ufixed_fast,double)

  FLOATDOUBLE_SC_X_FIXED_Y(float,sc_fixed)
  FLOATDOUBLE_SC_X_FIXED_Y(float,sc_ufixed)
  FLOATDOUBLE_SC_X_FIXED_Y(float,sc_fixed_fast)
  FLOATDOUBLE_SC_X_FIXED_Y(float,sc_ufixed_fast)
  FLOATDOUBLE_SC_X_FIXED_Y(double,sc_fixed)
  FLOATDOUBLE_SC_X_FIXED_Y(double,sc_ufixed)
  FLOATDOUBLE_SC_X_FIXED_Y(double,sc_fixed_fast)
  FLOATDOUBLE_SC_X_FIXED_Y(double,sc_ufixed_fast)
//END: SC FIXED CONVERSION TO float, double


//BEGIN SC_FIXED CONVERSION TO dummy objects
#define CONVERTER_SPECIALIZATION_SC_X_FIXED_Y_TO_DUMMY(T_WRITE,ID) \
template <int W,int I,sc_q_mode Q,sc_o_mode O, int N> \
class datconv<T_WRITE<W,I,Q,O,N>, impl::dummy<ID> >{ \
  \
  public: \
  \
  datconv(converterchannel_options<T_WRITE<W,I,Q,O,N>, impl::dummy<ID> >* options){};   \
  \
  impl::dummy<ID> conv(T_WRITE<W,I,Q,O,N> x) \
  { \
    ostringstream warn_str; \
    warn_str << "Conversion to impl::dummy is being performed, check your source code, returning default constructed dummy instance" << endl; \
    SC_REPORT_WARNING("ANDRES-Converterchannels",warn_str.str().c_str()); \
    return impl::dummy<ID>(); \
  };  \
};

CONVERTER_SPECIALIZATION_SC_X_FIXED_Y_TO_DUMMY(sc_fixed,1)
CONVERTER_SPECIALIZATION_SC_X_FIXED_Y_TO_DUMMY(sc_ufixed,1)
CONVERTER_SPECIALIZATION_SC_X_FIXED_Y_TO_DUMMY(sc_fixed_fast,1)
CONVERTER_SPECIALIZATION_SC_X_FIXED_Y_TO_DUMMY(sc_ufixed_fast,1)
CONVERTER_SPECIALIZATION_SC_X_FIXED_Y_TO_DUMMY(sc_fixed,2)
CONVERTER_SPECIALIZATION_SC_X_FIXED_Y_TO_DUMMY(sc_ufixed,2)
CONVERTER_SPECIALIZATION_SC_X_FIXED_Y_TO_DUMMY(sc_fixed_fast,2)
CONVERTER_SPECIALIZATION_SC_X_FIXED_Y_TO_DUMMY(sc_ufixed_fast,2)
//END SC_FIXED CONVERSION TO dummy objects

//BEGIN: SC FIXED VALUES TO SC BV/LV
  template <typename T_WRITE, typename T_READ, int N>             // intermediate class for conversion to sc_bv<N> / sc_lv<N>
  class datconv_fixed_scaling_blv_intermed : datconv_fixed_scaling_base<T_WRITE,T_READ>{

    public:

    datconv_fixed_scaling_blv_intermed(converterchannel_options<T_WRITE, T_READ>* options): datconv_fixed_scaling_base<T_WRITE,T_READ>(
      options->f2ipol,
      options->oflowpol,
      options->rangeScaling,
      options->rangeMin,
      options->rangeMax,
      options->rangeTargetMin,
      options->rangeTargetMax){}

    T_READ conv(T_WRITE x){
      T_READ tmp=static_cast<T_READ>(static_cast<sc_biguint<N> >(conv_base(x)));
      CONVERTERS_TRACE_CONV_DO(cerr<<"conv for "<<demangle<T_WRITE>()<<" --> "<<demangle<T_READ>()<<" has been called ("<<x<<" --> "<<tmp<<")"<<endl;);
      return tmp;
    }
  };
  #define SC_X_FIXED_Y_BLV(from,to) template <int W,int I,sc_q_mode Q,sc_o_mode O,int N,int N1> \
    class datconv<from<W,I,Q,O,N>, to<N1> > : public datconv_fixed_scaling_blv_intermed<from<W,I,Q,O,N>, to<N1> , N1>{\
      public:\
      datconv(converterchannel_options<from<W,I,Q,O,N>, to<N1> >* options):\
       datconv_fixed_scaling_blv_intermed<from<W,I,Q,O,N>, to<N1> , N1>(options){\
        CONVERTERS_TRACE_CONV_INSTANTIATED(#from"<W,I,Q,O,N>, "#to"<N1>",this); }\
    };

  SC_X_FIXED_Y_BLV(sc_fixed,sc_bv)
  SC_X_FIXED_Y_BLV(sc_ufixed,sc_bv)
  SC_X_FIXED_Y_BLV(sc_fixed_fast,sc_bv)
  SC_X_FIXED_Y_BLV(sc_ufixed_fast,sc_bv)
  SC_X_FIXED_Y_BLV(sc_fixed,sc_lv)
  SC_X_FIXED_Y_BLV(sc_ufixed,sc_lv)
  SC_X_FIXED_Y_BLV(sc_fixed_fast,sc_lv)
  SC_X_FIXED_Y_BLV(sc_ufixed_fast,sc_lv)
//END: SC FIXED VALUES TO SC BV/LV



#define SC_X_FIXED_Y_BLV(from,to) template <int W,int I,sc_q_mode Q,sc_o_mode O,int N,int N1> \
    class datconv<from<W,I,Q,O,N>, to<N1> > : public datconv_fixed_scaling_blv_intermed<from<W,I,Q,O,N>, to<N1> , N1>{\
      public:\
      datconv(converterchannel_options<from<W,I,Q,O,N>, to<N1> >* options):\
       datconv_fixed_scaling_blv_intermed<from<W,I,Q,O,N>, to<N1> , N1>(options){\
        CONVERTERS_TRACE_CONV_INSTANTIATED(#from"<W,I,Q,O,N>, "#to"<N1>",this); }\
    };

//BEGIN INT TO SC FIXED
// #define to sc_fixed
//   template <typename from,int W,int I,sc_q_mode Q,sc_o_mode O,int N>
//   class datconv<from, to<W,I,Q,O,N> > {
//     public:
//     datconv(converterchannel_options<from,to<W,I,Q,O,N> >* options):m_options(options) {
//       CONVERTERS_TRACE_CONV_INSTANTIATED("from -> sc_(u)fixed(_fast)",this);
//     }
//     to<W,I,Q,O,N> conv(from x){
//         to<W,I,Q,O,N> result;
//      std::cerr<<m_options->rangeMin<<","<<m_options->rangeMax<<"----"<<m_options->rangeTargetMin<<","<<m_options->rangeTargetMax<<"~~~~"<<m_options->rangeScaling<<std::endl;
//         CONVERTERS_TRACE_CONV_WORK_FIXED("T_WRITE,T_READ",from,"to",x,result);
//         return result;
//     }
//     private:
//       converterchannel_options<from,to<W,I,Q,O,N> >* m_options;
//   };
//
// #undef to

//END INT TO SC FIXED


#endif /* SC_INCLUDE_FX*/

} // end namespace TUV_ams_lib
// kate: space-indent on; indent-width 2; replace-tabs on;
